import logging
import os
import sys

"""
Utility functions
"""

# This is not currently used. We should improve our logging throughout.


def check_directory_exists_and_if_not_mkdir(directory, logger):
    """Checks if the given directory exists and creates it if it does not exist

    Parameters
    ----------
    directory: str
        Name of the directory

    Borrowed from bilby-pipe: https://git.ligo.org/lscsoft/bilby_pipe/
    """
    if not os.path.exists(directory):
        os.makedirs(directory)
        logger.debug(f"Making directory {directory}")
    else:
        logger.debug(f"Directory {directory} exists")


def setup_logger(outdir=None, label=None, log_level="INFO"):
    """Setup logging output: call at the start of the script to use

    Parameters
    ----------
    outdir, label: str
        If supplied, write the logging output to outdir/label.log
    log_level: str, optional
        ['debug', 'info', 'warning']
        Either a string from the list above, or an integer as specified
        in https://docs.python.org/2/library/logging.html#logging-levels

    Borrowed from bilby-pipe: https://git.ligo.org/lscsoft/bilby_pipe/
    """

    if "-v" in sys.argv or "--verbose" in sys.argv:
        log_level = "DEBUG"

    if isinstance(log_level, str):
        try:
            level = getattr(logging, log_level.upper())
        except AttributeError:
            raise ValueError(f"log_level {log_level} not understood")
    else:
        level = int(log_level)

    logger = logging.getLogger("dingo")
    logger.propagate = False
    logger.setLevel(level)

    streams = [isinstance(h, logging.StreamHandler) for h in logger.handlers]
    if len(streams) == 0 or not all(streams):
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(
            logging.Formatter(
                "%(asctime)s %(name)s %(levelname)-8s: %(message)s", datefmt="%H:%M"
            )
        )
        stream_handler.setLevel(level)
        logger.addHandler(stream_handler)

    if any([isinstance(h, logging.FileHandler) for h in logger.handlers]) is False:
        if label:
            if outdir:
                check_directory_exists_and_if_not_mkdir(outdir, logger)
            else:
                outdir = "."
            log_file = f"{outdir}/{label}.log"
            file_handler = logging.FileHandler(log_file)
            file_handler.setFormatter(
                logging.Formatter(
                    "%(asctime)s %(levelname)-8s: %(message)s", datefmt="%H:%M"
                )
            )

            file_handler.setLevel(level)
            logger.addHandler(file_handler)

    for handler in logger.handlers:
        handler.setLevel(level)


setup_logger()
logger = logging.getLogger("dingo")



"""
Step 1: Train unconditional nde
Step 2: Set up likelihood and prior
"""
from pathlib import Path

import yaml
from os import rename, makedirs
from os.path import dirname, join, isfile, exists
import argparse

from dingo.core.posterior_models import NormalizingFlowPosteriorModel
from dingo.gw.result import Result
from dingo.gw.inference.gw_samplers import GWSampler
from dingo.gw.importance_sampling.diagnostics import plot_diagnostics


def parse_args():
    parser = argparse.ArgumentParser(
        description="Importance sampling (IS) for dingo models."
    )
    parser.add_argument(
        "--settings",
        type=str,
        required=True,
        help="Path to settings file.",
    )
    parser.add_argument(
        "--outdir",
        type=str,
        default=None,
        help="Output directory for the unconditional nde and IS results.",
    )
    args = parser.parse_args()

    if args.outdir is None:
        args.outdir = dirname(args.settings)

    return args


def main():
    # parse args, load settings, load dingo parameter samples
    args = parse_args()
    with open(args.settings, "r") as fp:
        settings = yaml.safe_load(fp)
    try:
        result = Result(file_name=settings["parameter_samples"])
    except KeyError:
        # except statement for backward compatibility
        result = Result(file_name=settings["nde"]["data"]["parameter_samples"])
    metadata = result.settings
    samples = result.samples
    # for time marginalization, we drop geocent time from the samples
    inference_parameters = metadata["train_settings"]["data"][
        "inference_parameters"
    ].copy()
    time_marginalization_kwargs = settings.get("time_marginalization", None)
    time_marginalization = time_marginalization_kwargs is not None
    phase_marginalization_kwargs = settings.get("phase_marginalization", None)
    phase_marginalization = phase_marginalization_kwargs is not None
    calibration_marginalization_kwargs = settings.get(
        "calibration_marginalization", None
    )
    synthetic_phase_kwargs = settings.get("synthetic_phase", None)
    synthetic_phase = synthetic_phase_kwargs is not None
    # if sum([time_marginalization, phase_marginalization, synthetic_phase]) > 1:
    #    raise NotImplementedError(
    #        "Only one of time_marginalization, phase_marginalization and"
    #        "synthetic_phase can be set to True."
    #    )
    if time_marginalization and "geocent_time" in samples:
        if "geocent_time" in inference_parameters:
            samples.drop("geocent_time", axis=1, inplace=True)
            inference_parameters.remove("geocent_time")
    if phase_marginalization or synthetic_phase:
        if "phase" in inference_parameters:
            samples.drop("phase", axis=1, inplace=True)
            inference_parameters.remove("phase")
    if "nde" in settings:
        settings["nde"]["data"]["inference_parameters"] = inference_parameters
        settings["nde"]["data"]["parameters"] = inference_parameters
        # TODO: train_unconditional_density_estimator should not accept
        #  settings["data"]["parameters"], such that the line above can be removed.

    # Step 1: Build proposal distribution.
    #
    # We use the dingo posterior as our proposal distribution. We need to be able to
    # sample from, and evaluate this distribution. Here, we therefore train an
    # unconditional neural density estimator (nde) to recover the posterior density from
    # dingo samples. This typically required, since one loses the tractable likelihood
    # when using GNPE for dingo. This is not a big deal though, since we can cheaply
    # sample from the dingo posterior, such that one can easily train an unconditional
    # nde to recover the posterior density.

    if "log_prob" not in samples.columns:
        # Use GPS time as name for now.
        event_name = str(result.event_metadata["time_event"])
        nde_name = settings["nde"].get(
            "path", join(args.outdir, f"nde-{event_name}.pt")
        )
        if isfile(nde_name):
            print(f"Loading nde at {nde_name} for event {event_name}.")
            nde = NormalizingFlowPosteriorModel(
                model_filename=nde_name,
                device=settings["nde"]["training"]["device"],
                load_training_info=False,
            )
        else:
            print(f"Training new nde for event {event_name}.")
            nde = result.train_unconditional_flow(
                inference_parameters,
                settings["nde"],
                train_dir=args.outdir,
            )
            print(f"Renaming trained nde model to {nde_name}.")
            rename(join(args.outdir, "model_latest.pt"), nde_name)

        # Step 1a: Sample from proposal.
        nde_sampler = GWSampler(model=nde)
        nde_sampler.run_sampler(num_samples=settings["num_samples"])
        result = nde_sampler.to_result()

    # else:
    #     nde_sampler = GWSamplerUnconditional(
    #         result=result,
    #         synthetic_phase_kwargs=synthetic_phase_kwargs,
    #     )
    #
    # # Step 2: Sample from proposal.
    #
    # print(f'Generating {settings["num_samples"]} samples from proposal distribution.')
    # nde_sampler.run_sampler(num_samples=settings["num_samples"])

    # Step 2: Importance sample.
    #
    # Our target distribution is the posterior p(theta|d) = p(d|theta) * p(theta) / p(
    # d). For importance sampling, we need to evaluate the *unnormalized* posterior,
    # so we only need the likelihood p(d|theta) and the prior p(theta), but not the
    # evidence p(d).
    #
    # Sample from proposal distribution q(theta|d) and reweight the samples theta_i with
    #
    #       w_i = p(theta_i|d) / q(theta_i|d)
    #
    # to obtain weighted samples from the proposal distribution.

    # log_evidences = []
    # log_evidences_std = []
    # for idx in range(20):
    #     nde_sampler.run_sampler(num_samples=settings["num_samples"])
    #     nde_sampler.importance_sample(
    #         num_processes=settings.get("num_processes", 1),
    #         time_marginalization_kwargs=time_marginalization_kwargs,
    #         phase_marginalization_kwargs=phase_marginalization_kwargs,
    #     )
    #     log_evidences.append(nde_sampler.log_evidence)
    #     log_evidences_std.append(nde_sampler.log_evidence_std)
    # import numpy as np
    # log_evidences = np.array(log_evidences)
    # log_evidences_std = np.array(log_evidences_std)
    # print(np.std(log_evidences) / np.mean(log_evidences_std))

    if synthetic_phase:
        print(f"Sampling synthetic phase.")
        result.sample_synthetic_phase(synthetic_phase_kwargs)

    print(f"Importance sampling.")
    result.importance_sample(
        num_processes=settings.get("num_processes", 1),
        time_marginalization_kwargs=time_marginalization_kwargs,
        phase_marginalization_kwargs=phase_marginalization_kwargs,
        calibration_marginalization_kwargs=calibration_marginalization_kwargs,
    )
    result.print_summary()
    result.to_file(file_name=Path(args.outdir, "dingo_samples_weighted.hdf5"))

    # Diagnostics
    diagnostics_dir = join(args.outdir, "IS-diagnostics")
    if not exists(diagnostics_dir):
        makedirs(diagnostics_dir)
    print("Plotting diagnostics.")
    plot_diagnostics(
        result,
        diagnostics_dir,
        num_processes=settings.get("num_processes", 1),
        **settings.get("slice_plots", {}),
    )


if __name__ == "__main__":
    main()



"""Default settings for unconditional density estimation"""


def get_default_nde_settings_3d(
    device="cpu",
    num_workers=0,
    inference_parameters=None,
):
    settings = {
        "data": {
            "inference_parameters": inference_parameters,
        },
        "model": {
            "type": "nsf",
            "num_flow_steps": 5,
            "base_transform_kwargs": {
                "hidden_dim": 128,
                "num_transform_blocks": 2,
                "activation": "elu",
                "dropout_probability": 0.1,
                "batch_norm": True,
                "num_bins": 8,
                "base_transform_type": "rq-coupling",
            },
            "input_dim": 3,
            "context_dim": None,
        },
        "training": {
            "device": device,
            "num_workers": num_workers,
            "train_fraction": 0.9,
            "batch_size": 4096,
            "epochs": 10,
            "optimizer": {"type": "adam", "lr": 0.005},
            "scheduler": {"type": "cosine", "T_max": 10},
        },
    }
    return settings


DEFAULT_NDE_SETTINGS_2D = {
    "data": {
        "inference_parameters": ["GNPE:H1_time_proxy", "GNPE:L1_time_proxy"],
        "parameter_samples": None,
    },
    "model": {
        "type": "nsf",
        "num_flow_steps": 5,
        "base_transform_kwargs": {
            "hidden_dim": 64,
            "num_transform_blocks": 2,
            "activation": "elu",
            "dropout_probability": 0.1,
            "batch_norm": "true",
            "num_bins": 8,
            "base_transform_type": "rq-coupling",
        },
    },
    # nde training
    "training": {
        "device": "cpu",
        "num_workers": 0,
        "train_fraction": 0.9,
        "batch_size": 4096,
        "epochs": 10,
        "optimizer": {"type": "adam"},
        "lr": 0.001,
        "scheduler": {
            "type": "cosine",
            "T_max": 10,
        },
    },
}



import logging
from typing import Dict, List, Literal, Tuple
from dingo.core.utils.misc import get_version
import packaging.version as pv

import torch

_logger = logging.getLogger(__name__)
WINDOW_FACTOR_FIX_VERSION = pv.parse("0.9.0")

Device = Literal["meta", "cuda", "mps", "hip", "cpu"]


def torch_available_devices() -> List[Device]:
    """
    Returns a list of all available PyTorch devices,
    ordered: cuda, mps, hip, cpu
    Note: 'meta' is not included from the returned list,
    even if supported.

    Returns
    -------
    List of available device identifiers
    """
    devices: List[Device] = []

    # cuda
    if torch.cuda.is_available():
        devices.append("cuda")

    # mps
    try:
        if torch.backends.mps.is_available():
            devices.append("mps")
    except AttributeError:
        pass

    # hip
    try:
        if hasattr(torch, "hip") and torch.hip.is_available():
            devices.append("hip")
    except AttributeError:
        pass

    # cpu
    devices.append("cpu")

    return devices


def torch_load_with_fallback(
    filename: str, preferred_map_location: Device = "cuda"
) -> Tuple[Dict, torch.device]:
    """
    Loads a PyTorch file with fallback behavior:
    1. Tries preferred_map_location (default: cuda)
    2. Falls back to CUDA/MPS/HIP if available
    3. Finally falls back to CPU

    Returns
    -------
    Loaded model and torch device on which it has been loaded
    """

    try:
        r = (
            torch.load(filename, map_location=preferred_map_location),
            torch.device(preferred_map_location),
        )
        _logger.debug(f"loaded model {filename} to {preferred_map_location}")
        return r
    except (RuntimeError, AttributeError):
        # AttributeError can occur due to PyTorch bug: when CUDA is requested
        # on Mac, PyTorch's internal fallback tries torch.mps.current_device()
        # which doesn't exist (torch.mps lacks feature parity with torch.cuda)
        pass

    devices = torch_available_devices()

    for location in [d for d in devices if d != preferred_map_location]:
        try:
            r = torch.load(filename, map_location=location), torch.device(location)
            _logger.debug(
                f"loaded model {filename} to fallback device {location} "
                f"(preferred device was {preferred_map_location})"
            )
            return r
        except (RuntimeError, AttributeError):
            pass

    raise RuntimeError(
        f"failed to load model {filename} on any device, " "tried: {', '.join(devices)}"
    )


def check_minimum_version(version_str: str, raise_exception: bool = False) -> None:
    """
    Check that the version string is greater than a certain minimum value.

    By default, logs a warning. Optionally, raises an exception.

    This is used to handle major code changes that may break backwards compatibility
    with previously trained models or generated results.

    Parameters
    ----------
    version_str : str
        Version string to check, e.g., "version=0.8.5" or "0.8.5".

    raise_exception : bool
        If True, raise an exception if the version is below the minimum required version.
    """
    if "None" in version_str:
        version_str = "dingo=0.0.0"
    version_str = version_str.split("=", 1)[1]
    version = pv.parse(version_str)

    if version < WINDOW_FACTOR_FIX_VERSION:
        error_str = (
            f"This object was created using Dingo version {version} < {WINDOW_FACTOR_FIX_VERSION}, which broke backwards compatibility."
            f"\nFor models trained prior to this change, new inference results will be unreliable."
            f"\nPreviously-generated result files should be used with caution."
            f"\nReasons for backward compatibility breaking:\n"
            f"\nv{WINDOW_FACTOR_FIX_VERSION}: Change to window factor usage, see "
            f"https://git.ligo.org/pe/pe-group-coordination/-/issues/1#note_1469386."
        )
        if raise_exception:
            raise ValueError(error_str)
        else:
            _logger.warning("\n========\nWARNING!\n\n" + error_str + "\n=======\n")


def update_model_config(model_settings: dict):
    """
    Update the model settings to ensure backwards compatibility with networks
    trained using previous versions of Dingo.

    Parameters
    ----------
    model_settings: dict
        Model settings to be updated.
    """
    if model_settings.get("type") == "nsf+embedding":
        model_settings["posterior_model_type"] = "normalizing_flow"
        del model_settings["type"]
        model_settings["posterior_kwargs"] = model_settings["nsf_kwargs"]
        del model_settings["nsf_kwargs"]
        model_settings["embedding_kwargs"] = model_settings["embedding_net_kwargs"]
        del model_settings["embedding_net_kwargs"]



import os

from bilby_pipe.job_creation.nodes import AnalysisNode

from dingo.pipe.utils import _strip_unwanted_submission_keys


class SamplingNode(AnalysisNode):
    def __init__(self, inputs, generation_node, dag):
        super(AnalysisNode, self).__init__(inputs)
        self.dag = dag
        self.generation_node = generation_node
        self.request_cpus = inputs.request_cpus
        self.device = inputs.device

        data_label = generation_node.job_name
        base_name = data_label.replace("generation", "sampling")
        self.job_name = base_name
        self.label = self.job_name

        self.setup_arguments()


        # Add extra arguments for dingo
        self.arguments.add("label", self.label)
        self.arguments.add("event-data-file", generation_node.event_data_file)

        self.extra_lines.extend(self._checkpoint_submit_lines())
        env_vars = []
        # if self.request_cpus > 1:
        #     env_vars.append("OMP_NUM_THREADS=1")
        if getattr(self, "disable_hdf5_locking", None):
            env_vars.append("USE_HDF5_FILE_LOCKING=FALSE")
        if env_vars:
            self.extra_lines.append(f"environment = \"{' '.join(env_vars)}\"")

        for req in inputs.sampling_requirements:
            self.requirements.append(req)

        if self.device == "cuda":
            self.extra_lines.append("request_gpus = 1")


        if self.inputs.transfer_files or self.inputs.osg:
            input_files_to_transfer = [
                str(generation_node.event_data_file),
                str(self.inputs.complete_ini_file),
            ]
            if self.inputs.osg:
                sites = self.inputs.gpu_desired_sites
                if sites is not None:
                    self.extra_lines.append(f'MY.DESIRED_Sites = "{sites}"')
                self.requirements.append("IS_GLIDEIN=?=True")

                # modifying the paths for OSDF networks
                network_files = []
                for s in [self.inputs.model, self.inputs.model_init]:
                    if s is None:
                        continue
                    if "osdf" in s:
                        # stripping osdf prefix as it is not needed
                        network_files.append(f"igwn+osdf://{s.replace('/osdf', '')}")
                    else:
                        network_files.append(s)

                input_files_to_transfer.extend(network_files)

                if self.transfer_container:
                    input_files_to_transfer.append(self.inputs.container)

                # Credentials are needed to access any OSDF files
                if any(["osdf" in s for s in input_files_to_transfer]):
                    self.extra_lines.extend(self.scitoken_lines)

            self.extra_lines.extend(
                self._condor_file_transfer_lines(
                    input_files_to_transfer,
                    [self._relative_topdir(self.inputs.outdir, self.inputs.initialdir)],
                )
            )

            self.arguments.add("outdir", os.path.relpath(self.inputs.outdir))

        self.process_node()
        self.job.add_parent(generation_node.job)

        if self.inputs.simple_submission:
            _strip_unwanted_submission_keys(self.job)

    @property
    def executable(self):
        return self._get_executable_path("dingo_pipe_sampling")

    @property
    def samples_file(self):
        # TODO: Maybe remove -- not needed.
        return os.path.join(
            self.inputs.result_directory, self.label + ".hdf5"
        )

    @property
    def result_file(self):
        return self.samples_file



import argparse
from pathlib import Path

from dingo.gw.result import Result


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--outdir", type=str)
    parser.add_argument("--label", type=str)
    parser.add_argument("--extension", type=str)
    parser.add_argument("--result", type=str, nargs="+")
    parser.add_argument("--merge", action="store_true")

    args = parser.parse_args()

    if args.merge:
        print(f"Merging {len(args.result)} parts into complete Result.")
        sub_results = []
        for file_name in args.result:
            sub_results.append(Result(file_name=file_name))

        result = Result.merge(sub_results)
        result.print_summary()

        output_file = Path(args.outdir) / (args.label.replace("_merge", "") + ".hdf5")
        result.to_file(output_file)



import argparse
import copy
import textwrap
from multiprocessing import Pool
from pathlib import Path
from typing import Dict, Tuple
from functools import partial

import numpy as np
import pandas as pd
import yaml
from bilby.gw.prior import BBHPriorDict
from threadpoolctl import threadpool_limits
from torchvision.transforms import Compose

from dingo.gw.dataset.waveform_dataset import WaveformDataset
from dingo.gw.domains import build_domain
from dingo.gw.prior import build_prior_with_defaults
from dingo.gw.SVD import ApplySVD, SVDBasis
from dingo.gw.transforms import WhitenFixedASD
from dingo.gw.waveform_generator import (
    NewInterfaceWaveformGenerator,
    WaveformGenerator,
    generate_waveforms_parallel,
)
from dingo.core.utils.misc import call_func_strict_output_dim


def generate_parameters_and_polarizations(
    waveform_generator: WaveformGenerator,
    prior: BBHPriorDict,
    num_samples: int,
    num_processes: int,
) -> Tuple[pd.DataFrame, Dict[str, np.ndarray]]:
    """
    Generate a dataset of waveforms based on parameters drawn from the prior.

    Parameters
    ----------
    waveform_generator : WaveformGenerator
    prior : Prior
    num_samples : int
    num_processes : int

    Returns
    -------
    pandas DataFrame of parameters
    dictionary of numpy arrays corresponding to waveform polarizations
    """
    print("Generating dataset of size " + str(num_samples))
    parameters = pd.DataFrame(prior.sample(num_samples))

    if num_processes > 1:
        with threadpool_limits(limits=1, user_api="blas"):
            with Pool(processes=num_processes) as pool:
                polarizations = generate_waveforms_parallel(
                    waveform_generator, parameters, pool
                )
    else:
        polarizations = generate_waveforms_parallel(waveform_generator, parameters)

    # Find cases where waveform generation failed and only return data for successful ones
    wf_failed = np.any(np.isnan(polarizations["h_plus"]), axis=1)
    if wf_failed.any():
        idx_failed = np.where(wf_failed)[0]
        idx_ok = np.where(~wf_failed)[0]
        polarizations_ok = {k: v[idx_ok] for k, v in polarizations.items()}
        parameters_ok = parameters.iloc[idx_ok]
        failed_percent = 100 * len(idx_failed) / len(parameters)
        print(
            f"{len(idx_failed)} out of {len(parameters)} configuration ({failed_percent:.1f}%) failed to generate."
        )
        with pd.option_context("display.max_rows", None, "display.max_columns", None):
            print(parameters.iloc[idx_failed])
        print(
            f"Only returning the {len(idx_ok)} successfully generated configurations."
        )
        return parameters_ok, polarizations_ok

    return parameters, polarizations


def train_svd_basis(dataset: WaveformDataset, size: int, n_train: int):
    """
    Train (and optionally validate) an SVD basis.

    Parameters
    ----------
    dataset : WaveformDataset
        Contains waveforms to be used for building SVD.
    size : int
        Number of elements to keep for the SVD basis.
    n_train : int
        Number of training waveforms to use. Remaining are used for validation. Note
        that the actual number of training waveforms is n_train * len(polarizations),
        since there is one waveform used for each polarization.

    Returns
    -------
    SVDBasis, n_train, n_test
        Since EOB waveforms can fail to generate, provide also the number used in
        training and validation.
    """
    # Prepare data for training and validation.
    train_data = np.vstack([val[:n_train] for val in dataset.polarizations.values()])
    test_data = np.vstack([val[n_train:] for val in dataset.polarizations.values()])
    test_parameters = pd.concat(
        [
            # I would like to save the polarization, but saving the dataframe with
            # string columns causes problems. Fix this later.
            # dataset.parameters.iloc[n_train:].assign(polarization=pol)
            dataset.parameters.iloc[n_train:]
            for pol in dataset.polarizations
        ]
    )
    test_parameters.reset_index(drop=True, inplace=True)

    print("Building SVD basis.")
    basis = SVDBasis()
    basis.generate_basis(train_data, size)

    assert np.allclose(basis.V[: dataset.domain.min_idx], 0)

    # Since there is a possibility that the size of the dataset returned by
    # generate_parameters_and_polarizations is smaller than requested, we don't assume
    # that there are n_test samples. Instead we just look at the size of the test
    # dataset.
    if test_data.size != 0:
        basis.compute_test_mismatches(
            test_data, parameters=test_parameters, verbose=True
        )

    # Return also the true number of samples. Some EOB waveforms may have failed to
    # generate, so this could be smaller than the number requested.
    n_ifos = len(dataset.polarizations)
    n_train = len(train_data) // n_ifos
    n_test = len(test_data) // n_ifos

    return basis, n_train, n_test


def generate_dataset(settings: Dict, num_processes: int) -> WaveformDataset:
    """
    Generate a waveform dataset.

    Parameters
    ----------
    settings : dict
        Dictionary of settings to configure the dataset
    num_processes : int

    Returns
    -------
    A WaveformDataset based on the settings.
    """

    prior = build_prior_with_defaults(settings["intrinsic_prior"])
    domain = build_domain(settings["domain"])

    new_interface_flag = settings["waveform_generator"].get("new_interface", False)
    if new_interface_flag:
        waveform_generator = NewInterfaceWaveformGenerator(
            domain=domain,
            **settings["waveform_generator"],
        )
    else:
        waveform_generator = WaveformGenerator(
            domain=domain,
            **settings["waveform_generator"],
        )

    dataset_dict = {"settings": settings}

    if "compression" in settings:
        compression_transforms = []

        if "whitening" in settings["compression"]:
            compression_transforms.append(
                WhitenFixedASD(
                    domain,
                    asd_file=settings["compression"]["whitening"],
                    inverse=False,
                )
            )

        if "svd" in settings["compression"]:
            svd_settings = settings["compression"]["svd"]

            # Load an SVD basis from file, if specified.
            if "file" in svd_settings:
                basis = SVDBasis(file_name=svd_settings["file"])

            # Otherwise, generate the basis based on simulated waveforms.
            else:
                # If using whitened waveforms, then the SVD should be based on these.
                waveform_generator.transform = Compose(compression_transforms)

                n_train = svd_settings["num_training_samples"]
                n_test = svd_settings.get("num_validation_samples", 0)

                func = partial(
                    generate_parameters_and_polarizations,
                    waveform_generator,
                    prior,
                    num_processes=num_processes,
                )
                parameters, polarizations = call_func_strict_output_dim(
                    func, n_train + n_test
                )
                svd_dataset_settings = copy.deepcopy(settings)
                svd_dataset_settings["num_samples"] = len(parameters)
                del svd_dataset_settings["compression"]["svd"]

                # We build a WaveformDataset containing the SVD-training waveforms
                # because when constructed, it will automatically zero the waveforms
                # below f_min. This is useful for EOB waveforms, which are Fourier
                # transformed from time domain, and hence are nonzero below f_min. The
                # waveforms need to be zeroed below f_min because this corresponds to
                # setting the lower bound of the likelihood integral.

                svd_dataset = WaveformDataset(
                    dictionary={
                        "parameters": parameters,
                        "polarizations": polarizations,
                        "settings": svd_dataset_settings,
                    }
                )
                basis, n_train, n_test = train_svd_basis(
                    svd_dataset, svd_settings["size"], n_train
                )

            compression_transforms.append(ApplySVD(basis))
            dataset_dict["svd"] = basis.to_dictionary()

        waveform_generator.transform = Compose(compression_transforms)

    func = partial(
        generate_parameters_and_polarizations,
        waveform_generator,
        prior,
        num_processes=num_processes,
    )
    parameters, polarizations = call_func_strict_output_dim(
        func, settings["num_samples"]
    )
    dataset_dict["parameters"] = parameters
    dataset_dict["polarizations"] = polarizations

    dataset_dict[settings["num_samples"]] = len(parameters)
    dataset = WaveformDataset(dictionary=dataset_dict)
    return dataset


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(
            """\
        Generate a waveform dataset based on a settings file.
        """
        ),
    )
    parser.add_argument(
        "--settings_file",
        type=str,
        required=True,
        help="YAML file containing database settings",
    )
    parser.add_argument(
        "--num_processes",
        type=int,
        default=1,
        help="Number of processes to use in pool for parallel waveform generation",
    )
    parser.add_argument(
        "--out_file",
        type=str,
        default="waveform_dataset.hdf5",
        help="Name of file for storing dataset.",
    )
    return parser.parse_args()


def _generate_dataset_main(
    settings_file: str, out_file: str, num_processes: int
) -> None:
    if not Path(settings_file).is_file():
        raise FileNotFoundError(f"dataset generation, failed to find {settings_file}")
    if not Path(out_file).parent.is_dir():
        raise FileNotFoundError(
            f"dataset generation: can not create {out_file}: "
            f"{Path(out_file).parent} does not exist"
        )
    # Load settings
    with open(settings_file, "r") as f:
        settings = yaml.safe_load(f)

    dataset = generate_dataset(settings, num_processes)
    dataset.to_file(str(out_file))


def main() -> None:
    args = parse_args()
    _generate_dataset_main(args.settings_file, args.out_file, args.num_processes)


if __name__ == "__main__":
    main()



from typing import Optional
import numpy as np
from copy import deepcopy
from scipy.signal.windows import tukey
from scipy.interpolate import interp1d

from bilby.gw.detector import PowerSpectralDensity

from dingo.gw.prior import default_extrinsic_dict
from dingo.gw.prior import BBHExtrinsicPriorDict


def get_window(window_kwargs):
    """Compute window from window_kwargs."""
    type = window_kwargs["type"]
    if type == "tukey":
        roll_off, T, f_s = (
            window_kwargs["roll_off"],
            window_kwargs["T"],
            window_kwargs["f_s"],
        )
        alpha = 2 * roll_off / T
        w = tukey(int(T * f_s), alpha)
        return w
    else:
        raise NotImplementedError(f"Unknown window type {type}.")


def get_extrinsic_prior_dict(extrinsic_prior):
    """Build dict for extrinsic prior by starting with
    default_extrinsic_dict, and overwriting every element for which
    extrinsic_prior is not default.
    TODO: Move to dingo.gw.prior.py?"""
    extrinsic_prior_dict = default_extrinsic_dict.copy()
    for k, v in extrinsic_prior.items():
        if v.lower() != "default":
            extrinsic_prior_dict[k] = v
    return extrinsic_prior_dict


def get_mismatch(a, b, domain, asd_file=None):
    """
    Mistmatch is 1 - overlap, where overlap is defined by
    inner(a, b) / sqrt(inner(a, a) * inner(b, b)).
    See e.g. Eq. (44) in https://arxiv.org/pdf/1106.1021.pdf.

    Parameters
    ----------
    a
    b
    domain
    asd_file

    Returns
    -------

    """
    if asd_file is not None:
        # whiten a and b, such that we can use flat-spectrum inner products below
        psd = PowerSpectralDensity(asd_file=asd_file)
        asd_interp = interp1d(
            psd.frequency_array, psd.asd_array, bounds_error=False, fill_value=np.inf
        )
        asd_array = asd_interp(domain.sample_frequencies)
        a = a / asd_array
        b = b / asd_array
    min_idx = domain.min_idx
    inner_ab = np.sum((a.conj() * b)[..., min_idx:], axis=-1).real
    inner_aa = np.sum((a.conj() * a)[..., min_idx:], axis=-1).real
    inner_bb = np.sum((b.conj() * b)[..., min_idx:], axis=-1).real
    overlap = inner_ab / np.sqrt(inner_aa * inner_bb)
    return 1 - overlap


def get_standardization_dict(
    extrinsic_prior_dict, wfd, selected_parameters, transform=None
):
    """
    Calculates the mean and standard deviation of parameters. This is needed for
    standardizing neural-network input and output.

    Parameters
    ----------
    extrinsic_prior_dict : dict
    wfd : WaveformDataset
    selected_parameters : list[str]
        List of parameters for which to estimate standardization factors.
    transform : Transform
        Operator that will generate samples for parameters contained in
        selected_parameters that are not contained in the intrinsic or extrinsic prior.
        (E.g., H1_time, L1_time_proxy)

    Returns
    -------

    """
    # The intrinsic standardization is estimated based on the entire dataset.
    mean_intrinsic, std_intrinsic = wfd.parameter_mean_std()

    # Some of the extrinsic prior parameters have analytic means and standard
    # deviations. If possible, this will either get these, or else it will estimate
    # them numerically.
    ext_prior = BBHExtrinsicPriorDict(extrinsic_prior_dict)
    mean_extrinsic, std_extrinsic = ext_prior.mean_std(ext_prior.keys())

    # Check that overlap between intrinsic and extrinsic parameters is only
    # due to fiducial values (-> std 0)
    for k in std_intrinsic.keys() & std_extrinsic.keys():
        if std_intrinsic[k] != 0:
            raise ValueError(
                f"Expected intrinsic prior for {k} to be a fixed value in the waveform dataset, "
                f"since {k} is specified as an extrinsic prior in the train settings and will be sampled"
                f"during training. However, the standard deviation of {k} is non-zero: {std_intrinsic[k]}"
                f"Please re-generate the waveform dataset with a fixed value for {k}."
            )

    # Merge dicts, overwriting fiducial values for parameters (e.g.,
    # luminosity_distance) in intrinsic parameters by the extrinsic ones
    mean = {**mean_intrinsic, **mean_extrinsic}
    std = {**std_intrinsic, **std_extrinsic}

    # For all remaining parameters that require standardization, we use the transform
    # to sample these and estimate the mean and standard deviation numerically.
    additional_parameters = [p for p in selected_parameters if p not in mean]
    if additional_parameters:
        num_samples = min(100_000, len(wfd.parameters))
        samples = {p: np.empty(num_samples) for p in additional_parameters}
        for n in range(num_samples):
            sample = {"parameters": wfd.parameters.iloc[n].to_dict()}
            sample = transform(sample)
            for p in additional_parameters:
                # This assumes all of the additional parameters are contained within
                # extrinsic_parameters. We have set it up so this is the case for the
                # GNPE proxies and the detector coalescence times.
                samples[p][n] = sample["extrinsic_parameters"][p]
        mean_additional = {p: np.mean(samples[p]).item() for p in additional_parameters}
        std_additional = {p: np.std(samples[p]).item() for p in additional_parameters}

        mean.update(mean_additional)
        std.update(std_additional)

    standardization_dict = {
        "mean": {k: mean[k] for k in selected_parameters},
        "std": {k: std[k] for k in selected_parameters},
    }
    return standardization_dict



class UnpackDict(object):
    """
    Unpacks the dictionary to prepare it for final output of the dataloader.
    Only returns elements specified in selected_keys.
    """
    def __init__(self, selected_keys):
        self.selected_keys = selected_keys

    def __call__(self, input_sample):
        return [input_sample[k] for k in self.selected_keys]


import copy

import numpy as np
from scipy import stats
from dingo.gw.noise.synthetic.asd_parameterization import fit_broadband_noise
from dingo.gw.noise.asd_dataset import ASDDataset
from dingo.gw.noise.synthetic.utils import (
    get_index_for_elem,
)


class KDE:
    """
    Kernel Density Estimation (KDE) class for sampling ASDs.
    """
    def __init__(self, parameter_dict, sampling_settings):
        """
        Parameters
        ----------
        parameter_dict : dict
            Dictionary containing the parameters of the ASDs used for fitting the synthetic distribution.
        sampling_settings : dict
            Dictionary containing the settings for the sampling.

        """

        self.parameter_dicts = parameter_dict
        self.settings = sampling_settings

        detectors = parameter_dict.keys()
        self.spectral_kde = {det: [] for det in detectors}
        self.broadband_kde = {det: [] for det in detectors}

    def fit(self, weights=None):
        """
        Fit the KDEs to the parameters saved in 'self.parameter_dict'.
        Parameters
        ----------
        weights : array_like, optional
            Weights for the KDEs. If None, all weights are set to 1.
        """

        for det, param_dict in self.parameter_dicts.items():
            y_values = param_dict["y_values"]
            x_positions = param_dict["x_positions"]
            spectral_features = param_dict["spectral_features"]

            for i in range(spectral_features.shape[1]):
                try:
                    spectral_kde = stats.gaussian_kde(
                        spectral_features[:, i, :].T,
                        bw_method=float(self.settings["bandwidth_spectral"]),
                        weights=weights,
                    )
                except np.linalg.LinAlgError:
                    print(
                        "Warning: Singular Matrix encountered in spectral KDE. Adding small Gaussian noise..."
                    )
                    perturbed_features = spectral_features[:, i, :] + np.random.normal(
                        0, 1.0e-4, size=spectral_features[:, i, :].shape
                    )
                    spectral_kde = stats.gaussian_kde(
                        perturbed_features.T,
                        bw_method=float(self.settings["bandwidth_spectral"]),
                        weights=weights,
                    )
                self.spectral_kde[det].append(spectral_kde)

            split_indices = [0, len(x_positions)]
            split_indices += [
                get_index_for_elem(x_positions, f) + 1
                for f in self.settings["split_frequencies"]
            ]
            split_indices = sorted(split_indices)

            for i in range(len(split_indices) - 1):
                vals = y_values[:, split_indices[i]: split_indices[i + 1]].T
                kde_vals = stats.gaussian_kde(
                    vals, bw_method=float(self.settings["bandwidth_spline"])
                )
                self.broadband_kde[det].append(kde_vals)

    def sample(self, num_samples, rescaling_ys=None):
        """
        Sample a synthetic ASD dataset from the fitted KDEs
        ----------
        Parameters:
        num_samples (int): Number of samples to draw.
        rescaling_ys (dict): Optional dictionary of spline y-values used for rescaling the base noise.
        """

        parameters_dicts = copy.deepcopy(self.parameter_dicts)
        asds_dict = {}

        for det, param_dict in self.parameter_dicts.items():
            asds_dict[det] = []

            features = np.zeros((num_samples, len(self.spectral_kde[det]), 3))
            for i, kde in enumerate(self.spectral_kde[det]):
                features[:, i, :] = kde.resample(size=num_samples).T

            parameters_dicts[det]["spectral_features"] = features

            y_values = np.zeros((num_samples, 0))

            for i, kde in enumerate(self.broadband_kde[det]):
                y_values = np.concatenate(
                    (y_values, kde.resample(size=num_samples).T), axis=1
                )
            # rescale base noise
            if rescaling_ys:
                y_values_mean = np.mean(y_values, axis=0)
                y_values = (
                    y_values - y_values_mean[None, :] + rescaling_ys[det]
                )
            parameters_dicts[det]["y_values"] = y_values

        return parameters_dicts


def get_rescaling_params(filenames, parameterization_settings):
    """
    Get the parameters of the ASDs that are used for rescaling.
    Parameters
    ----------
    filenames : dict
        Dictionary containing the paths to the ASD files.
    parameterization_settings : dict
        Dictionary containing the settings for the parameterization.
    """
    parameters = {}
    for det, asd_path in filenames.items():
        asd_dataset = ASDDataset(asd_path)
        psd = asd_dataset.asds[det][0] ** 2

        # we only need the parameterized broadband noise
        _, ys = fit_broadband_noise(
            domain=asd_dataset.domain,
            psd=np.log(psd),
            num_spline_positions=parameterization_settings["num_spline_positions"],
            sigma=float(parameterization_settings["sigma"]),
        )
        parameters[det] = ys

    return parameters



from bilby_pipe.job_creation.nodes import MergeNode as BilbyMergeNode

from dingo.pipe.utils import _strip_unwanted_submission_keys


class MergeNode(BilbyMergeNode):

    run_node_on_osg = False
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs, detectors=[])
        if self.inputs.simple_submission:
            _strip_unwanted_submission_keys(self.job)

    @property
    def executable(self):
        return self._get_executable_path("dingo_result")

    @property
    def result_file(self):
        label = self.label.replace("_merge", "")
        return f"{self.inputs.result_directory}/{label}.hdf5"



from typing import Dict, List, Optional, Union
import h5py
import numpy as np
import torch.utils.data
from torchvision.transforms import Compose

from dingo.core.dataset import DingoDataset, recursive_hdf5_load
from dingo.gw.SVD import SVDBasis, ApplySVD
from dingo.gw.domains import build_domain
from dingo.gw.transforms import WhitenFixedASD


class WaveformDataset(DingoDataset, torch.utils.data.Dataset):
    """This class stores a dataset of waveforms (polarizations) and corresponding
    parameters.

    It can load the dataset either from an HDF5 file or suitable dictionary.

    It is possible to either load the entire dataset into memory or to load the dataset during training
    (leave_waveforms_on_disk=True) to reduce the memory footprint.
    At the moment, it is only possible to load the waveforms on-demand since the
    standardization dict for all parameters in the dataset has to be computed at the
    beginning of training.

    The waveform data is consumed through a __getitem__() or __getitems__() call which optionally loads the
    polarizations and applies a chain of transformations, which are classes that implement a __call__() method.
    """

    dataset_type = "waveform_dataset"

    def __init__(
        self,
        file_name: Optional[str] = None,
        dictionary: Optional[dict] = None,
        transform=None,
        precision: Optional[str] = None,
        domain_update: Optional[dict] = None,
        svd_size_update: Optional[int] = None,
        leave_waveforms_on_disk: Optional[bool] = False,
    ):
        """
        For constructing, provide either file_name, or dictionary containing data and
        settings entries, or neither.

        Parameters
        ----------
        file_name : str
            HDF5 file containing a dataset
        dictionary : dict
            Contains settings and data entries. The dictionary keys should be
            'settings', 'parameters', and 'polarizations'.
        transform : Transform
            Transform to be applied to dataset samples when accessed through __getitem__
        precision : str ('single', 'double')
            If provided, changes precision of loaded dataset.
        domain_update : dict
            If provided, update domain from existing domain using new settings.
        svd_size_update : int
            If provided, reduces the SVD size when decompressing (for speed).
        leave_waveforms_on_disk : bool
            If True, the values for the waveforms are not loaded into RAM when initializing the
            waveform dataset. Instead, they are loaded lazily in __getitem__().
        """
        self.domain = None
        self.transform = transform
        self.decompression_transform = None
        self.file_handle = None
        self.precision = precision

        if leave_waveforms_on_disk:
            leave_on_disk_keys = ["polarizations"]
        else:
            leave_on_disk_keys = []
        super().__init__(
            file_name=file_name,
            dictionary=dictionary,
            data_keys=["parameters", "polarizations", "svd"],
            leave_on_disk_keys=leave_on_disk_keys,
        )
        self.file_name = file_name

        if self.settings is not None:
            self.load_supplemental(domain_update, svd_size_update)
            self.svd_size_update = svd_size_update

    def load_supplemental(
        self,
        domain_update: Optional[dict] = None,
        svd_size_update: Optional[int] = None,
    ):
        """Method called immediately after loading a dataset.

        Creates (and possibly updates) domain, updates dtypes, and initializes any
        decompression transform. Also zeros data below f_min, and truncates above f_max.

        Parameters
        ----------
        domain_update : dict
            If provided, update domain from existing domain using new settings.
        svd_size_update : int
            If provided, reduces the SVD size when decompressing (for speed).
        """
        self.domain = build_domain(self.settings["domain"])

        # We always call update_domain() (even if domain_update is None) because we
        # want to be sure that the data are consistent with the saved settings. In
        # particular, this zeroes the waveforms for f < f_min.
        self.update_domain(domain_update)

        # Update dtypes if necessary
        if self.precision is not None:
            if self.parameters is not None:
                self.parameters = self.parameters.astype(self.real_type, copy=False)
            if self.polarizations is not None:
                for k, v in self.polarizations.items():
                    self.polarizations[k] = v.astype(self.complex_type, copy=False)

            # This should probably be moved to the SVDBasis class.
            if self.svd is not None:
                self.svd["V"] = self.svd["V"].astype(self.complex_type, copy=False)
                # For backward compatibility; in future, this will be there.
                if "s" in self.svd:
                    self.svd["s"] = self.svd["s"].astype(self.real_type, copy=False)

        if self.settings.get("compression", None) is not None:
            self.initialize_decompression(svd_size_update)

    def update_domain(self, domain_update: Optional[dict] = None):
        """
        Update the domain based on new configuration.

        The waveform dataset provides waveform polarizations in a particular domain. In
        Frequency domain, this is [0, domain._f_max]. Furthermore, data is set to 0 below
        domain._f_min. In practice one may want to train a network based on  slightly
        different domain settings, which corresponds to truncating the likelihood
        integral.

        This method provides functionality for that. It truncates and/or zeroes the
        dataset to the range specified by the domain, by calling domain.update_data.

        Parameters
        ----------
        domain_update : dict
            Settings dictionary. Must contain a subset of the keys contained in
            domain_dict.
        """
        if domain_update is not None:
            self.domain.update(domain_update)

        # Determine where any domain adjustment must be applied. If the dataset is SVD
        # compressed, then adjust the SVD matrices. Otherwise, adjust the dataset
        # itself (if it already has been loaded).
        if (
            self.settings.get("compression", None) is not None
            and "svd" in self.settings["compression"]
        ):
            self.svd["V"] = self.domain.update_data(self.svd["V"], axis=0)
        elif self.polarizations is not None:
            for k, v in self.polarizations.items():
                self.polarizations[k] = self.domain.update_data(v)

    def initialize_decompression(self, svd_size_update: Optional[int] = None):
        """
        Sets up decompression transforms. These are applied to the raw dataset before
        self.transform. E.g., SVD decompression.

        Parameters
        ----------
        svd_size_update : int
            If provided, reduces the SVD size when decompressing (for speed).
        """
        decompression_transform_list = []

        # These transforms must be in reverse order compared to when dataset was
        # constructed.

        if "svd" in self.settings["compression"]:
            assert self.svd is not None

            # We allow the option to reduce the size of the SVD used for decompression,
            # since decompression is the costliest preprocessing operation. Be careful
            # when using this to not introduce a large mismatch.
            if svd_size_update is not None:
                if svd_size_update > self.svd["V"].shape[-1] or svd_size_update < 0:
                    raise ValueError(
                        f"Cannot truncate SVD from size "
                        f"{self.svd['V'].shape[-1]} to size "
                        f"{svd_size_update}."
                    )
                self.svd["V"] = self.svd["V"][:, :svd_size_update]
                self.svd["s"] = self.svd["s"][:svd_size_update]
                if self.polarizations is not None:
                    for k, v in self.polarizations.items():
                        self.polarizations[k] = v[:, :svd_size_update]

            svd_basis = SVDBasis(dictionary=self.svd)
            decompression_transform_list.append(ApplySVD(svd_basis, inverse=True))

        if "whitening" in self.settings["compression"]:
            decompression_transform_list.append(
                WhitenFixedASD(
                    self.domain,
                    asd_file=self.settings["compression"]["whitening"],
                    inverse=True,
                    precision=self.precision,
                )
            )

        self.decompression_transform = Compose(decompression_transform_list)

    @property
    def real_type(self):
        if self.precision is not None:
            if self.precision == "single":
                return np.float32
            elif self.precision == "double":
                return np.float64
            else:
                raise TypeError(
                    "Precision can only be changed to 'single' or 'double'."
                )
        else:
            return None

    @property
    def complex_type(self):
        if self.precision is not None:
            if self.precision == "single":
                return np.complex64
            elif self.precision == "double":
                return np.complex128
            else:
                raise TypeError(
                    "Precision can only be changed to 'single' or 'double'."
                )
        else:
            return None

    def __len__(self):
        """The number of waveform samples."""
        return len(self.parameters)

    def __getitem__(self, idx: int) -> Dict[str, Dict[str, Union[float, np.ndarray]]]:
        """
        Return a nested dictionary containing parameters and waveform polarizations
        for samples with indices `idx_batched`. If defined, a chain of transformations is applied to
        the waveform data.

        Parameters
        ----------
        idx : int
            Index of the sample in the WaveformDataset to return.

        Returns
        -------
        Dict[str, Dict[str, Union[float, np.ndarray]]]
            Nested dictionary containing parameters and waveform polarizations.
        """
        return self.__getitems__([idx])[0]

    def __getitems__(
        self, batched_idx: list[int]
    ) -> list[Dict[str, Dict[str, Union[float, np.ndarray]]]]:
        """
        Return a nested dictionary containing parameters and waveform polarizations
        for sample with index `idx`. If defined, a chain of transformations is applied to
        the waveform data.

        Parameters
        ----------
        batched_idx : list[int]
            List of indices to return.

        Returns
        -------
        repackaged_data : list[Dict[str, Dict[str, Union[float, np.ndarray]]]]
            Nested dictionary containing parameters and waveform polarizations.
        """

        # Get parameters and data for idx
        if "polarizations" in self._leave_on_disk_keys:
            # Load polarizations from disk
            if self.file_handle is None:
                # Open hdf5 file
                self.file_handle = h5py.File(self.file_name, "r")

            polarizations = recursive_hdf5_load(
                self.file_handle, keys=self._leave_on_disk_keys, idx=batched_idx
            )["polarizations"]
            # Apply domain update to set waveform to zero for f < f_min
            if self.svd is None:
                polarizations = {
                    pol: self.domain.update_data(waveforms)
                    for pol, waveforms in polarizations.items()
                }
            parameters = {
                k: v if isinstance(v, float) else v.to_numpy()
                for k, v in self.parameters.iloc[batched_idx].items()
            }
            # Convert parameters to dict
            if not isinstance(parameters, dict):
                parameters = parameters.to_dict()
            # Update precision
            if self.precision is not None:
                polarizations = {
                    k: v.astype(self.complex_type, copy=False)
                    for k, v in polarizations.items()
                }
            # Perform SVD size update on waveform
            if self.svd is not None and self.svd_size_update is not None:
                for k, v in polarizations.items():
                    if len(v.shape) == 1:
                        polarizations[k] = v[: self.svd_size_update]
                    else:
                        polarizations[k] = v[:, : self.svd_size_update]
        else:
            parameters = {
                k: v.to_numpy() for k, v in self.parameters.iloc[batched_idx].items()
            }
            polarizations = {
                pol: waveforms[batched_idx]
                for pol, waveforms in self.polarizations.items()
            }

        # Decompression transforms are assumed to apply only to the waveform,
        # and do not involve parameters.
        if self.decompression_transform is not None:
            polarizations = self.decompression_transform(polarizations)

        # Main transforms can depend also on parameters.
        data = {"parameters": parameters, "waveform": polarizations}
        if self.transform is not None:
            data = self.transform(data)

        # The DataLoader expects a list of items from the dataset, which it will later
        # collate. However, depending on self.transform, here data is either a nested
        # dict of arrays or a list of arrays, each array having length batch_size.
        # Repackage data into a list of length batch_size, each item having the same
        # structure as before.

        if isinstance(data, dict):
            data = [
                {k1: {k2: v2[j] for k2, v2 in v1.items()} for k1, v1 in data.items()}
                for j in range(len(batched_idx))
            ]
        elif isinstance(data, list):
            data = [
                [data[i][j] for i in range(len(data))] for j in range(len(batched_idx))
            ]
        else:
            raise NotImplementedError()
        return data

    def __del__(self):
        # Close hdf5 file when wfd is deleted
        if self.file_handle is not None:
            self.file_handle.close()

    def parameter_mean_std(self):
        mean = self.parameters.mean().to_dict()
        std = self.parameters.std().to_dict()
        return mean, std



import argparse

import numpy as np
import torch
import yaml

from dingo.core.utils.backward_compatibility import torch_load_with_fallback


def append_stage():

    parser = argparse.ArgumentParser()
    parser.add_argument("--checkpoint", type=str, required=True)
    parser.add_argument("--stage_settings_file", type=str, required=True)
    parser.add_argument("--out_file", type=str, required=True)
    parser.add_argument("--replace", type=int)
    args = parser.parse_args()

    # trying to load on CUDA, MPS, HIP or CPU
    d, _ = torch_load_with_fallback(args.checkpoint)

    stages = [
        v
        for k, v in d["metadata"]["train_settings"]["training"].items()
        if k.startswith("stage_")
    ]
    num_stages = len(stages)
    print(f"Checkpoint training plan consists of {num_stages} stages.")

    with open(args.stage_settings_file, "r") as f:
        new_stage = yaml.safe_load(f)

    if args.replace is not None:
        if args.replace < 0 or args.replace >= num_stages:
            raise ValueError(
                f"Invalid argument replace={args.replace}. Valid values "
                f"are {list(range(num_stages))}."
            )
        current_epoch = d["epoch"]
        stage_epoch = np.sum([s["epochs"] for s in stages[: args.replace]])
        if current_epoch > stage_epoch:
            print(
                f"WARNING: Modification to training plan changes a training stage "
                f"that has already started. Current model epoch is {current_epoch}. "
                f"Proceed at your own risk!"
            )
        print(f"Replacing planned stage {args.replace} with new stage.")
        new_stage_number = args.replace
    else:
        print(f"Appending new stage to training plan.")
        new_stage_number = num_stages

    d["metadata"]["train_settings"]["training"][f"stage_{new_stage_number}"] = new_stage
    print("Summary of new training plan:")
    print(
        yaml.dump(
            d["metadata"]["train_settings"]["training"],
            default_flow_style=False,
            sort_keys=False,
        )
    )

    torch.save(d, args.out_file)



import numpy as np
import pytest
from scipy.interpolate import interp1d

from dingo.gw.domains import MultibandedFrequencyDomain
from dingo.gw.gwutils import get_mismatch
from dingo.gw.prior import build_prior_with_defaults
from dingo.gw.waveform_generator import WaveformGenerator, NewInterfaceWaveformGenerator


@pytest.fixture
def mfd():
    domain_settings = {
        "nodes": [20.0, 26.0, 34.0, 46.0, 62.0, 78.0, 1038.0],
        "delta_f_initial": 0.0625,
        "base_domain": {
            "type": "UniformFrequencyDomain",
            "f_min": 20.0,
            "f_max": 1037.9375,
            "delta_f": 0.0625,
        },
    }
    domain = MultibandedFrequencyDomain(**domain_settings)
    return domain


@pytest.fixture(params=["IMRPhenomXPHM", "SEOBNRv4PHM", "SEOBNRv5PHM", "SEOBNRv5HM"])
def approximant(request):
    return request.param


@pytest.fixture
def intrinsic_prior(approximant):
    intrinsic_dict = {
        "mass_1": "bilby.core.prior.Constraint(minimum=10.0, maximum=80.0)",
        "mass_2": "bilby.core.prior.Constraint(minimum=10.0, maximum=80.0)",
        "mass_ratio": "bilby.gw.prior.UniformInComponentsMassRatio(minimum=0.125, maximum=1.0)",
        "chirp_mass": "bilby.gw.prior.UniformInComponentsChirpMass(minimum=25.0, maximum=100.0)",
        "luminosity_distance": 1000.0,
        "theta_jn": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
        "phase": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
        "a_1": "bilby.core.prior.Uniform(minimum=0.0, maximum=0.99)",
        "a_2": "bilby.core.prior.Uniform(minimum=0.0, maximum=0.99)",
        "tilt_1": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
        "tilt_2": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
        "phi_12": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
        "phi_jl": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
        "geocent_time": "bilby.core.prior.Uniform(minimum=-0.1, maximum=0.1)",
    }
    prior = build_prior_with_defaults(intrinsic_dict)
    return prior


@pytest.fixture
def wfg_mfd(mfd, approximant):
    if approximant in ["SEOBNRv5PHM", "SEOBNRv5HM"]:
        wfg_class = NewInterfaceWaveformGenerator
    else:
        wfg_class = WaveformGenerator
    return wfg_class(
        approximant=approximant,
        domain=mfd,
        f_ref=10.0,
        f_start=10.0,
        spin_conversion_phase=0.0,
    )


@pytest.fixture
def wfg_ufd(mfd, approximant):
    if approximant in ["SEOBNRv5PHM", "SEOBNRv5HM"]:
        wfg_class = NewInterfaceWaveformGenerator
    else:
        wfg_class = WaveformGenerator
    return wfg_class(
        approximant=approximant,
        domain=mfd.base_domain,
        f_ref=10.0,
        f_start=10.0,
        spin_conversion_phase=0.0,
    )


@pytest.fixture
def num_evaluations(approximant):
    if approximant == "SEOBNRv4PHM":
        return 1
    else:
        return 10


@pytest.fixture
def tolerances(approximant):
    # Return max mismatches in MFD, UFD.
    if approximant == "IMRPhenomXPHM":
        return 1e-4, 1e-3
    else:
        return 1e-9, 1e-3


# Uncomment to test only one approximant.
try:
    import pyseobnr

    approximant_list = ["IMRPhenomXPHM", "SEOBNRv4PHM", "SEOBNRv5PHM"]
except ImportError:
    approximant_list = ["IMRPhenomXPHM", "SEOBNRv4PHM"]


@pytest.mark.parametrize("approximant", approximant_list)
def test_decimation(
    intrinsic_prior, wfg_mfd, wfg_ufd, mfd, num_evaluations, tolerances
):
    mismatches_mfd = []
    mismatches_ufd = []
    for _ in range(num_evaluations):
        p = intrinsic_prior.sample()

        wf_mfd = wfg_mfd.generate_hplus_hcross(p)
        wf_ufd = wfg_ufd.generate_hplus_hcross(p)

        # Compare UFD waveforms decimated to MFD against waveforms generated in MFD.
        # FD-native waveforms (e.g., Phenom) can be generated directly in MFD, so this
        # comparison is non-exact. TD-native waveforms (e.g., EOB) are first generated
        # in FD and decimated to MFD within the WaveformGenerator class, so this should
        # be exact.

        wf_ufd_decimated = {k: mfd.decimate(v) for k, v in wf_ufd.items()}

        mismatches_mfd.append(
            [
                get_mismatch(
                    wf_mfd[pol],
                    wf_ufd_decimated[pol],
                    mfd,
                    asd_file="aLIGO_ZERO_DET_high_P_asd.txt",
                )
                for pol in ["h_plus", "h_cross"]
            ]
        )

        # Also compare UFD waveforms against MFD waveforms interpolated to UFD. This
        # comparison is always going to be non-exact.

        ufd = mfd.base_domain
        wf_mfd_interpolated = {
            k: ufd.update_data(interp1d(mfd(), v, fill_value="extrapolate")(ufd()))
            for k, v in wf_mfd.items()
        }

        mismatches_ufd.append(
            [
                get_mismatch(
                    wf_ufd[pol],
                    wf_mfd_interpolated[pol],
                    ufd,
                    asd_file="aLIGO_ZERO_DET_high_P_asd.txt",
                )
                for pol in ["h_plus", "h_cross"]
            ]
        )

    mismatches_mfd = np.array(mismatches_mfd)
    mismatches_ufd = np.array(mismatches_ufd)

    assert np.max(mismatches_mfd) < tolerances[0]
    assert np.max(mismatches_ufd) < tolerances[1]


@pytest.mark.parametrize("approximant", approximant_list)
def test_decimation_m(
    intrinsic_prior, wfg_mfd, wfg_ufd, mfd, num_evaluations, tolerances
):
    mismatches_mfd = []
    mismatches_ufd = []
    for _ in range(num_evaluations):
        p = intrinsic_prior.sample()

        wf_mfd = wfg_mfd.generate_hplus_hcross_m(p)
        wf_ufd = wfg_ufd.generate_hplus_hcross_m(p)
        modes = wf_mfd.keys()

        # Compare UFD waveforms decimated to MFD against waveforms generated in MFD.
        # FD-native waveforms (e.g., Phenom) can be generated directly in MFD, so this
        # comparison is non-exact. TD-native waveforms (e.g., EOB) are first generated
        # in FD and decimated to MFD within the WaveformGenerator class, so this should
        # be exact.

        wf_ufd_decimated = {
            k: {k2: mfd.decimate(v2) for k2, v2 in v.items()} for k, v in wf_ufd.items()
        }

        mismatches_mfd.append(
            [
                [
                    get_mismatch(
                        wf_mfd[m][pol],
                        wf_ufd_decimated[m][pol],
                        mfd,
                        asd_file="aLIGO_ZERO_DET_high_P_asd.txt",
                    )
                    for pol in ["h_plus", "h_cross"]
                ]
                for m in modes
            ]
        )

        # Also compare UFD waveforms against MFD waveforms interpolated to UFD. This
        # comparison is always going to be non-exact.

        ufd = mfd.base_domain
        wf_mfd_interpolated = {
            k: {
                k2: ufd.update_data(
                    interp1d(mfd(), v2, fill_value="extrapolate")(ufd())
                )
                for k2, v2 in v.items()
            }
            for k, v in wf_mfd.items()
        }

        mismatches_ufd.append(
            [
                [
                    get_mismatch(
                        wf_ufd[m][pol],
                        wf_mfd_interpolated[m][pol],
                        ufd,
                        asd_file="aLIGO_ZERO_DET_high_P_asd.txt",
                    )
                    for pol in ["h_plus", "h_cross"]
                ]
                for m in modes
            ]
        )

    mismatches_mfd = np.array(mismatches_mfd)
    mismatches_ufd = np.array(mismatches_ufd)

    assert np.max(mismatches_mfd) < tolerances[0]

    # Some of the negative m modes do not do well, so we exclude by taking the median.
    assert np.median(mismatches_ufd) < 10 * tolerances[1]



from typing import List, Optional, Union
import ast
import h5py
import numpy as np
import pandas as pd

from dingo.core.utils.misc import get_version


def recursive_hdf5_save(group, d):
    for k, v in d.items():
        if v is None:
            continue
        elif isinstance(v, dict):
            next_group = group.create_group(k)
            recursive_hdf5_save(next_group, v)
        elif isinstance(v, np.ndarray):
            group.create_dataset(k, data=v)
        elif isinstance(v, pd.DataFrame):
            group.create_dataset(k, data=v.to_records(index=False))
        elif isinstance(v, (int, float, complex, str, list)):
            # TODO: Set scalars as attributes?
            group.create_dataset(k, data=v)
        else:
            raise TypeError(f"Cannot save datatype {type(v)} as hdf5 dataset.")


def recursive_hdf5_load(
    group,
    keys: Optional[List[str]] = None,
    idx: Optional[Union[int, List[int]]] = None,
):
    """This is a generic helper function to recursively load data from an HDF5 file.

    Parameters
    ----------
    group: h5py.Group
        Group from which to recursively load data.
    keys: list[str] or None
        List of keys to load. If None, load all keys.
    idx: int or list[int] or None
        If idx is provided, only the datapoints corresponding to the given indices are loaded.
    """
    d = {}
    for k, v in group.items():
        if keys is None or k in keys:
            if isinstance(v, h5py.Group):
                d[k] = recursive_hdf5_load(v, idx=idx)
            else:
                # Load values from hdf5 file as np.ndarray
                if idx is None:
                    # Load all values
                    d[k] = v[...]
                elif isinstance(idx, list) and len(idx) > 1:
                    # Load batch of indices: hdf5 load requires index list to be sorted
                    sorting = np.argsort(idx)
                    sorted_idx = np.array(idx)[sorting]
                    reverse_sorting = np.zeros_like(sorting)
                    reverse_sorting[sorting] = np.arange(len(sorting))
                    d[k] = v[sorted_idx][reverse_sorting]
                else:
                    # Load specific idx
                    d[k] = v[idx]
                # Update data types
                # If the array has column names, load it as a pandas DataFrame
                if d[k].dtype.names is not None:
                    d[k] = pd.DataFrame(d[k])
                # Convert 0-dimensional arrays to scalars
                elif d[k].ndim == 0:
                    d[k] = d[k].item()
                    if isinstance(d[k], bytes):
                        # Assume this is a string.
                        d[k] = d[k].decode()
                # If an array is 1D and of type object, assume it originated as a list
                # of strings.
                elif d[k].ndim == 1 and d[k].dtype == "O":
                    d[k] = [x.decode() for x in d[k]]
    return d


class DingoDataset:
    """This is a generic dataset class with save / load methods.

    A common use case is to inherit multiply from DingoDataset and
    torch.utils.data.Dataset, in which case the subclass picks up these I/O methods,
    and DingoDataset is acting as a Mixin class.

    Alternatively, if the torch Dataset is not needed, then DingoDataset can be
    subclassed directly.
    """

    dataset_type = "dingo_dataset"

    def __init__(
        self,
        file_name: Optional[str] = None,
        dictionary: Optional[dict] = None,
        data_keys: Optional[List] = None,
        leave_on_disk_keys: Optional[list] = None,
    ):
        """
        For constructing, provide either file_name, or dictionary containing data and
        settings entries, or neither.

        Parameters
        ----------
        file_name : str
            HDF5 file containing a dataset
        dictionary : dict
            Contains settings and data entries. The data keys should be the same as
            save_keys
        data_keys : list
            Variables that should be saved / loaded. This allows for class to store
            additional variables beyond those that are saved. Typically, this list
            would be provided by any subclass.
        leave_on_disk_keys: Optional[list]
            Keys for which the values are not loaded into RAM when initializing the dataset.
            This reduces the memory footprint during training. Instead, the values are
            loaded from the HDF5 file during training.
        """
        self._data_keys = list(data_keys)  # Make a copy before modifying.
        self._data_keys.append("version")

        # Ensure all potential variables have None values to begin
        for key in self._data_keys:
            vars(self)[key] = None
        self.settings = None
        if leave_on_disk_keys is None:
            leave_on_disk_keys = []
        self._leave_on_disk_keys = leave_on_disk_keys

        # If data provided, load it
        if file_name is not None:
            self.from_file(file_name)
        elif dictionary is not None:
            self.from_dictionary(dictionary)

    def to_file(self, file_name: str, mode: str = "w"):
        print("Saving dataset to " + str(file_name))
        save_dict = {
            k: v
            for k, v in vars(self).items()
            if k in self._data_keys and v is not None
        }
        with h5py.File(file_name, mode) as f:
            recursive_hdf5_save(f, save_dict)
            if self.settings:
                f.attrs["settings"] = str(self.settings)
            if self.dataset_type:
                f.attrs["dataset_type"] = self.dataset_type

    def from_file(self, file_name: str):
        print(f"Loading dataset from {str(file_name)}.")
        if self._leave_on_disk_keys:
            print(f"Omitting data keys {self._leave_on_disk_keys}.")

        with h5py.File(file_name, "r") as f:
            loaded_dict = recursive_hdf5_load(
                f,
                keys=[k for k in self._data_keys if k not in self._leave_on_disk_keys],
            )
            # Set the keys that the class expects
            for k, v in loaded_dict.items():
                assert k in self._data_keys
                setattr(self, k, v)
            try:
                self.settings = ast.literal_eval(f.attrs["settings"])
            except KeyError:
                self.settings = None  # Is this necessary?

    def to_dictionary(self):
        dictionary = {
            k: v
            for k, v in vars(self).items()
            if (k in self._data_keys or k == "settings") and v is not None
        }
        return dictionary

    def from_dictionary(self, dictionary: dict):
        for k, v in dictionary.items():
            if k in self._data_keys or k == "settings":
                setattr(self, k, v)
        if "version" not in dictionary:
            self.version = f"dingo={get_version()}"



from typing import Tuple


def get_batch_size_of_input_sample(input_sample: dict) -> Tuple[bool, int]:
    """
    Parameters
    ----------
    input_sample : dict
        The input sample to transform. Can contain keys of
        parameters, polarizations

    Returns
    -------
    Tuple[bool, int]
        A tuple containing a boolean indicating if the input sample is batched
        and an integer representing the batch size of the input sample

    """
    if "parameters" in input_sample:
        params = next(iter(input_sample["parameters"].values()))

        # if parameters is a float not a batched input
        if isinstance(params, float):
            return False, 1
        else:
            # if single value array, return batch size 1
            if params.ndim < 1:
                return True, 1
            else:
                return True, params.shape[0]
    else:
        raise NotImplementedError(
            """The parameter key is used to determine the batch size of the
            input sample. If you want to determine the batch size a different
            way, you can implement it here.""" )



from bilby_pipe.bilbyargparser import BilbyArgParser
from bilby_pipe.utils import get_command_line_arguments, logger, parse_args

from dingo.gw.result import Result

logger.name = "dingo_pipe"


def create_parser():
    """Generate a parser for the plot script

    Returns
    -------
    parser: BilbyArgParser
        A parser with all the default options already added

    """
    parser = BilbyArgParser(ignore_unknown_config_file_keys=True)
    parser.add("--result", type=str, required=True, help="The result file")
    parser.add("--label", type=str, default="")
    # parser.add("--calibration", action="store_true", help="Generate calibration plot")
    parser.add("--corner", action="store_true", help="Generate corner plot")
    parser.add(
        "--weights", action="store_true", help="Generate plot of importance " "weights"
    )
    parser.add(
        "--log_probs",
        action="store_true",
        help="Generate plot of target" "versus proposal log probability",
    )
    # parser.add("--marginal", action="store_true", help="Generate marginal plots")
    # parser.add("--skymap", action="store_true", help="Generate skymap")
    # parser.add("--waveform", action="store_true", help="Generate waveform")
    parser.add(
        "--outdir", type=str, required=False, help="The directory to save the plots in"
    )
    parser.add(
        "--format",
        type=str,
        default="png",
        help="Format for making bilby_pipe plots, can be [png, pdf, html]. "
        "If specified format is not supported, will default to png.",
    )
    return parser


def main():
    args, _ = parse_args(get_command_line_arguments(), create_parser())

    logger.info(f"Generating plots for results file {args.result}")

    result = Result(file_name=args.result)

    if args.corner:
        logger.info("Generating corner plot.")
        result.plot_corner(filename=f"{args.outdir}/{args.label}_corner.pdf")

    if args.weights:
        logger.info("Generating weights plot.")
        result.plot_weights(filename=f"{args.outdir}/{args.label}_weights.png")

    if args.log_probs:
        logger.info("Generating log probability plot.")
        result.plot_log_probs(filename=f"{args.outdir}/{args.label}_log_probs.png")



from functools import partial

from multiprocessing import Pool
import numpy as np
import scipy
import tqdm
import scipy.optimize
from threadpoolctl import threadpool_limits

from dingo.gw.noise.synthetic.utils import get_index_for_elem, lorentzian_eval

P0_A, P0_Q = 5, 100
MIN_A, MIN_Q = 0, 10
MAX_A, MAX_Q = 12, 1000
MAXFEV = 5000000


def parameterize_asd_dataset(real_dataset, parameterization_settings, num_processes, verbose):
    """
    Parameterize a dataset of ASDs using a spline fit to the broadband noise and Lorentzians for the spectral features.

    Parameters
    ----------
    real_dataset : ASDDataset
        Dataset containing the ASDs to be parameterized.
    parameterization_settings : dict
        Dictionary containing the settings for the parameterization.
    num_processes : int
        Number of processes to use for parallelization.
    verbose : bool
        If True, print progress bars.
    """
    real_asd_dict = real_dataset.asds
    domain = real_dataset.domain

    parameters_dict = {det: [] for det in real_asd_dict.keys()}
    for det, asds in real_asd_dict.items():
        if num_processes > 1:
            with threadpool_limits(limits=1, user_api="blas"):
                with Pool(processes=num_processes) as pool:
                    parameters_dict[det] = parameterize_asds_parallel(
                        asds, domain, parameterization_settings, pool=pool, verbose=verbose
                    )
        else:
            parameters_dict[det] = parameterize_asds_parallel(asds, domain, parameterization_settings)

    return parameters_dict


def parameterize_asds_parallel(asds, domain, parameterization_settings, pool=None, verbose=False):
    """
    Helper function to be called for parallel ASD parameterization.

    Parameters
    ----------
    asds : array_like
        Array containing the ASDs to be parameterized.
    domain : Domain
        Domain object containing the frequency grid.
    parameterization_settings : dict
        Dictionary containing the settings for the parameterization.
    pool : Pool, optional
        Pool object for parallelization. If None, the function is not parallelized.
    verbose : bool
        If True, print progress bars.

    """

    task_func = partial(parameterize_single_psd, domain=domain, parameterization_settings=parameterization_settings)
    psds = asds ** 2
    if pool is not None:
        parameters_list = list(tqdm.tqdm(pool.imap(task_func, psds), total=psds.shape[0], disable=not verbose))
    else:
        parameters_list = list(tqdm.tqdm(map(task_func, psds), total=psds.shape[0], disable=not verbose))

    parameters = {
        feature: np.stack([asd_param[feature] for asd_param in parameters_list])
        for feature in parameters_list[0].keys()
    }
    # the x-positions are the same for all asds
    parameters["x_positions"] = parameters["x_positions"][0]

    return parameters


def parameterize_single_psd(real_psd, domain, parameterization_settings):
    """
    Parameterize a single ASD using a spline fit to the broadband noise and Lorentzians for the spectral features.

    Parameters
    ----------
    real_psd : array_like
        PSD to be parameterized.
    domain : Domain
        Domain object containing the frequency grid.
    parameterization_settings : dict
        Dictionary containing the settings for the parameterization.

    """
    sigma = float(parameterization_settings["sigma"])
    if not len(real_psd) == len(domain.sample_frequencies):
        real_psd = domain.update_data(real_psd)

    # optional truncation parameter for Lorentzians. Set to None if non-positive value is passed
    delta_f = parameterization_settings.get("delta_f", -1)
    delta_f = float(delta_f) if delta_f > 0 else None

    # transform psd to log space
    transformed_psd = np.log(real_psd)

    # parameterize broad-band noise
    xs, ys = fit_broadband_noise(
        domain=domain,
        psd=transformed_psd,
        num_spline_positions=parameterization_settings["num_spline_positions"],
        sigma=sigma,
    )
    spline = scipy.interpolate.CubicSpline(xs, ys)
    broadband_noise = spline(domain.sample_frequencies)

    _, features = fit_spectral(
        frequencies=domain.sample_frequencies,
        psd=transformed_psd,
        broadband_noise=broadband_noise,
        num_spectral_segments=parameterization_settings["num_spectral_segments"],
        sigma=sigma,
        delta_f=delta_f,
    )
    # features = np.random.normal(100, 10, size=(400, 3))
    parameter_dict = {"x_positions": xs, "y_values": ys, "spectral_features": features}

    return parameter_dict


def fit_broadband_noise(domain, psd, num_spline_positions, sigma, f_min=20):
    """
    Fit a spline to the broadband noise of a PSD.

    Parameters
    ----------
    domain : Domain
        Domain object containing the frequency grid.
    psd : array_like
        PSD to be parameterized.
    num_spline_positions : int
        Number of spline positions.
    sigma : float
        Standard deviation of the Gaussian noise used for the spline fit.
    f_min : float, optional
        position of the first node for the spline fi

    """
    frequencies = domain.sample_frequencies

    # standardize frequencies to the interval [0,1]
    standardized_frequencies = (frequencies - frequencies[0]) / (
        frequencies[-1] - frequencies[0]
    )

    # log-distributed x-positions in [20/f_max, 1] for the interpolating base noise spline
    log_xs = np.logspace(np.log10(f_min / domain.f_max), 0, num_spline_positions)

    # get indices corresponding to log_xs

    xs_indices = np.array(
        [get_index_for_elem(standardized_frequencies, log_x) for log_x in log_xs]
    )
    xs = frequencies[xs_indices]
    ys = []

    # for the y-values take the mean over an e-interval around the given x-value. Furthermore, we remove outliers
    # (further than 3 stds away from median) before computing the means such that spectral features are not considered
    # for the interpolation
    ind_min_old = 0
    for i, ind in enumerate(xs_indices):
        if i == 0:
            ind_min, ind_max = ind, int((ind + xs_indices[i + 1]) / 2)
        elif i == len(xs_indices) - 1:
            ind_min, ind_max = int((ind + xs_indices[i - 1]) / 2), ind + 1
        else:
            ind_min, ind_max = int((ind + xs_indices[i - 1]) / 2), int(
                (ind + xs_indices[i + 1]) / 2
            )

        # Apply filter to remove outliers, i.e. spectral lines
        assert ind_min != ind_max
        data = psd[ind_min:ind_max]
        mov_median = np.median(psd[ind_min_old:ind_max])

        # Any samples more than 3 noise stds away are removed as outliers
        ind_red = np.where(data - mov_median < 3 * sigma)
        data_red = data[ind_red]

        # Use mean over cleaned data as an estimate for the spline value (mean of the Gaussian noise)
        ys.append(np.mean(data_red))

        ind_min_old = ind_min

    return xs, np.array(ys)


def fit_spectral(
    frequencies, psd, broadband_noise, num_spectral_segments, sigma, delta_f
):
    """
    Fit Lorentzians to the spectral features of a PSD.

    Parameters
    ----------
    frequencies : array_like
        Frequency grid.
    psd : array_like
        PSD to be parameterized.
    broadband_noise : array_like
        Broadband noise of the PSD.
    num_spectral_segments : int
        Number of spectral segments.
    sigma : float
        Standard deviation of the Gaussian noise used for the spline fit.
    delta_f : float
        Truncation parameter for Lorentzians. Set to None if non-positive value is passed.

    """

    # divide frequency spectrum into equi-length sub-intervals, in each of which a single spectral line is fitted
    spectral_segments = np.array_split(
        np.arange(psd.shape[0]),
        num_spectral_segments,
    )

    spectral_features = np.array([[0.0, 0.0, 0.0]] * num_spectral_segments)
    lorentzians = np.array([])

    for i, segment in enumerate(spectral_segments):
        psd_data = psd[segment]
        frequency_data = frequencies[segment]
        broadband_noise_data = broadband_noise[segment]

        data = {
            "psd": psd_data,
            "broadband_noise": broadband_noise_data,
            "frequencies": frequency_data,
            "lower_freq": frequency_data[0],
            "upper_freq": frequency_data[-1],
        }

        with threadpool_limits(limits=1, user_api="blas"):
            try:
                popt = curve_fit(data, sigma, delta_f)
            except RuntimeError:
                popt = [0.0, 0.0, 0.0]  # will be replaced by sampled version below

        f0, A, Q = popt

        lorentzian = lorentzian_eval(frequency_data, f0, A, Q, delta_f=delta_f)
        # if no spectral line has been found -> peak smaller than 3 stds.
        # We don't want to fit random fluctuations as spectral lines
        if np.max(lorentzian) <= 3 * sigma:
            # sample from fixed distribution to group the event "no spectral line" together
            f0 = np.random.normal(frequency_data[0], 1.0e-3)
            A = np.random.normal(-1, 1.0e-8)
            Q = np.random.normal(500, 1)
            lorentzian = np.zeros_like(frequency_data)

        spectral_features[i] = [f0, A, Q]
        lorentzians = np.concatenate((lorentzians, lorentzian), axis=0)

    return lorentzians, spectral_features


def curve_fit(data, std, delta_f=None):
    """
    Fit a Lorentzian to the PSD.

    Parameters
    ----------
    data : dict
        Dictionary containing the PSD, broadband noise, and frequency grid.
    std : float
        Standard deviation of the Gaussian noise.
    delta_f : float
        Truncation parameter for Lorentzians. Set to None if non-positive value is passed.

    """
    func = partial(lorentzian_eval, delta_f=delta_f)

    popt, pcov = scipy.optimize.curve_fit(
        func,
        data["frequencies"],
        data["psd"] - data["broadband_noise"],
        p0=[(data["lower_freq"] + data["upper_freq"]) / 2, P0_A, P0_Q],
        sigma=[std] * len(data["frequencies"]),
        bounds=[[data["lower_freq"], MIN_A, MIN_Q], [data["upper_freq"], MAX_A, MAX_Q]],
        maxfev=MAXFEV,
    )
    return popt



import os
import torch
import numpy as np
import h5py
import json
import argparse


def parse_args():
    parser = argparse.ArgumentParser(
        description="Convert the weights of a trained Dingo model from a PyTorch pickle .pt file to HDF5,"
        " for distribution in the LVK's CVMFS.",
        epilog="Training history (optimizer_state_dict) is discarded.")
    parser.add_argument("-i", "--in_file", type=str, required=True,
            help='Input model ".pt" weights file')
    parser.add_argument("-o", "--out_file", type=str, required=True,
            help='Output model ".hdf5" weights file')
    parser.add_argument("-n", "--model_version_number", type=int, required=True,
            help="Model version number (integer). "
            "Will be included in the output filename and metadata.")
    return parser.parse_args()


def main():
    args = parse_args()
    if os.path.splitext(args.in_file)[-1] != '.pt':
        raise ValueError('Expected a .pt input file')
    if os.path.splitext(args.out_file)[-1] != '.hdf5':
        raise ValueError('Expected a .hdf5 output file')

    # Build output filename with the version number for this network
    # This is required for use on CVMFS
    root, ext = os.path.splitext(args.out_file)
    out_file_name = f'{root}_v{args.model_version_number}{ext}'
    print('Output will be written to', out_file_name)

    # Load data into CPU memory since we'll be saving it using CPU libraries
    d = torch.load(args.in_file, map_location=torch.device("cpu"))

    # Collect the names of the dicts that can be serialized to JSON; model_state_dict and
    # optimizer_state_dict contain torch.tensors and cannot be JSON serialized
    # In addition, we drop dicts related to training information that is not needed at inference time
    dicts_to_serialize = ['model_kwargs', 'epoch', 'metadata']


    with h5py.File(out_file_name, 'w') as f:
        # Save small nested dicts as json
        grp = f.create_group('serialized_dicts')
        for k in dicts_to_serialize:
            dict_str = json.dumps(d[k])
            grp.create_dataset(k, data=dict_str)

        # Save the OrderedDict containing the model weights
        # The keys are ordered alphanumerically as well.
        grp_model = f.create_group('model_weights')
        for k, v in d['model_state_dict'].items():
            if len(v.size()) > 0:
                grp_model.create_dataset(k, data=v.numpy(), fletcher32=True)
            else:
                # fletcher32 is not available for scalars
                grp_model.create_dataset(k, data=v.numpy())

        # Note we do not save optimizer_state_dict which is
        # not needed at inference time and saves a significant
        # amount of disk space.

        # Metadata for CVMFS LVK distribution
        # This needs to be exactly the same as the "basename" of the hdf5 file
        f.attrs['CANONICAL_FILE_BASENAME'] = os.path.basename(out_file_name)

        # Add a few metadata entries as attributes
        f.attrs['approximant'] = d['metadata']['dataset_settings']['waveform_generator']['approximant']
        f.attrs['epoch'] = d['epoch']
        # Add the dingo version used for training
        f.attrs['version'] = str(d.get('version'))
        # Make it clear to dingo_ls that this is file contains model weights
        f.attrs['dataset_type'] = 'trained_model' 


if __name__ == "__main__":
    main()




from .base import Domain
from .uniform_frequency_domain import UniformFrequencyDomain
from .time_domain import TimeDomain
from .multibanded_frequency_domain import MultibandedFrequencyDomain
from .build_domain import *



import argparse

import numpy as np
import yaml
from scipy.interpolate import interp1d

from dingo.gw.dataset import generate_parameters_and_polarizations
from dingo.gw.domains import build_domain, MultibandedFrequencyDomain
from dingo.gw.gwutils import get_mismatch
from dingo.gw.prior import build_prior_with_defaults
from dingo.gw.waveform_generator import (
    NewInterfaceWaveformGenerator,
    WaveformGenerator,
    generate_waveforms_parallel,
)


def _evaluate_multibanding_main(
    settings_file: str,
    num_samples: int,
):
    with open(settings_file, "r") as f:
        settings = yaml.safe_load(f)

    # Ignore any compression settings
    if "compression" in settings:
        del settings["compression"]

    # Update prior to challenge the multi-banding:
    #
    # (a) Set geocent_time = 0.12 s (boundary of usual prior + Earth-radius crossing time)
    # (b) Set chirp mass to bottom end of prior.
    prior = build_prior_with_defaults(settings["intrinsic_prior"])
    settings["intrinsic_prior"]["geocent_time"] = 0.12
    settings["intrinsic_prior"]["chirp_mass"] = prior["chirp_mass"].minimum
    # Rebuild prior with updated settings.
    prior = build_prior_with_defaults(settings["intrinsic_prior"])
    print("Prior")
    for k, v in prior.items():
        print(f"{k}: {v}")

    domain = build_domain(settings["domain"])
    print("\nDomain")
    print(domain.domain_dict)

    if not isinstance(domain, MultibandedFrequencyDomain):
        raise ValueError("Waveform dataset domain not a MultibandedFrequencyDomain.")

    if settings["waveform_generator"].get("new_interface", False):
        waveform_generator_mfd = NewInterfaceWaveformGenerator(
            domain=domain,
            **settings["waveform_generator"],
        )
        waveform_generator_ufd = NewInterfaceWaveformGenerator(
            domain=domain.base_domain,
            **settings["waveform_generator"],
        )
    else:
        waveform_generator_mfd = WaveformGenerator(
            domain=domain,
            **settings["waveform_generator"],
        )
        waveform_generator_ufd = WaveformGenerator(
            domain=domain.base_domain,
            **settings["waveform_generator"],
        )

    # Generate MFD waveforms.
    parameters, polarizations_mfd = generate_parameters_and_polarizations(
        waveform_generator_mfd, prior, num_samples, 1
    )

    # Generate UFD waveforms, re-using the parameter choices from before.
    polarizations_ufd = generate_waveforms_parallel(waveform_generator_ufd, parameters)

    # Compare UFD waveforms against MFD waveforms interpolated to MFD.
    mismatches = {}
    ufd = domain.base_domain
    mfd = domain
    for pol, d in polarizations_mfd.items():
        mismatches[pol] = np.empty(len(d))
        for i in range(len(d)):
            mfd_interpolated = interp1d(mfd(), d[i], fill_value="extrapolate")(ufd())
            mismatches[pol][i] = get_mismatch(
                polarizations_ufd[pol][i],
                mfd_interpolated,
                ufd,
                asd_file="aLIGO_ZERO_DET_high_P_asd.txt",
            )

    print("\nMismatches between UFD waveforms and MFD waveforms interpolated to MFD.")
    print(
        "This is a conservative estimate of the MFD performance when training "
        "networks."
    )
    mismatches = np.concatenate([v for v in mismatches.values()])
    print(f"num_samples = {num_samples}")
    print("  Mean mismatch = {}".format(np.mean(mismatches)))
    print("  Standard deviation = {}".format(np.std(mismatches)))
    print("  Max mismatch = {}".format(np.max(mismatches)))
    print("  Median mismatch = {}".format(np.median(mismatches)))
    print("  Percentiles:")
    print("    99    -> {}".format(np.percentile(mismatches, 99)))
    print("    99.9  -> {}".format(np.percentile(mismatches, 99.9)))
    print("    99.99 -> {}".format(np.percentile(mismatches, 99.99)))


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="Evaluate performance of multibanding on waveform dataset.",
    )
    parser.add_argument(
        "--settings-file",
        type=str,
        required=True,
        help="YAML file containing database settings",
    )
    parser.add_argument(
        "--num-samples",
        type=int,
        default=5000,
        help="Number of waveform evaluations for comparison.",
    )
    return parser.parse_args()


def main() -> None:
    args = parse_args()
    _evaluate_multibanding_main(args.settings_file, args.num_samples)



import torch

from dingo.gw.domains import build_domain, MultibandedFrequencyDomain
import pytest
import numpy as np


@pytest.fixture
def mfd_params():
    domain_settings = {
        "nodes": [20.0, 26.0, 34.0, 46.0, 62.0, 78.0, 1038.0],
        "delta_f_initial": 0.0625,
        "base_domain": {
            "type": "UniformFrequencyDomain",
            "f_min": 20.0,
            "f_max": 2048.0,
            "delta_f": 0.0625,
        },
    }
    return domain_settings


@pytest.fixture
def mfd(mfd_params):
    return MultibandedFrequencyDomain(**mfd_params)


def test_mfd_len(mfd_params, mfd):
    p = mfd_params
    len_expected = 0
    nodes = p["nodes"]
    for n in range(1, len(nodes)):
        len_expected += int(
            (nodes[n] - nodes[n - 1]) / p["delta_f_initial"] / 2 ** (n - 1)
        )
    assert len(mfd) == len_expected
    assert len(mfd()) == len_expected


def test_mfd(mfd_params, mfd):
    p = mfd_params
    nodes = p["nodes"]
    num_bins_cum = 0
    for n in range(1, len(nodes)):
        num_bins = int((nodes[n] - nodes[n - 1]) / p["delta_f_initial"] / 2 ** (n - 1))
        offset = (2 ** (n - 1) - 1) * p["delta_f_initial"] / 2
        frequencies_expected = (
            np.linspace(nodes[n - 1], nodes[n], num_bins, endpoint=False) + offset
        )
        frequencies = mfd()[num_bins_cum : num_bins_cum + num_bins]
        assert np.allclose(frequencies_expected, frequencies)
        num_bins_cum += num_bins
    assert mfd.min_idx == 0
    assert mfd.max_idx == len(mfd) - 1
    assert np.all(mfd.sample_frequencies_torch.numpy() == mfd.sample_frequencies)


def test_mfd_mask(mfd):
    mask = mfd.frequency_mask
    assert len(mask) == len(mfd)
    assert np.all(mask == 1.0)


def test_mfd_domain_dict(mfd):
    domain2 = build_domain(mfd.domain_dict)
    # dicts contain arrays so use np.testing.assert_equal to compare.
    np.testing.assert_equal(mfd.__dict__, domain2.__dict__)


def test_mfd_set_new_range(mfd_params, mfd):
    domain = mfd
    # test that ValueErrors are raised for infeasible inputs
    with pytest.raises(ValueError):
        domain._set_new_range(domain.f_max + 10, None)
    with pytest.raises(ValueError):
        domain._set_new_range(domain.f_min - 10, None)
    with pytest.raises(ValueError):
        domain._set_new_range(None, domain.f_max + 10)
    with pytest.raises(ValueError):
        domain._set_new_range(None, domain.f_min - 10)
    with pytest.raises(ValueError):
        domain._set_new_range(domain.f_min + 10, domain.f_min + 5)
    with pytest.raises(ValueError):
        domain._set_new_range(domain.f_min + 0.1, None)
    with pytest.raises(ValueError):
        domain._set_new_range(None, domain.f_max - 0.1)
    # test that setting new frequency range works as intended
    f_min_new, f_max_new = 40, 800
    domain._set_new_range(f_min_new, f_max_new)
    # Test that the end nodes are reasonable.
    assert domain.nodes[0] >= f_min_new
    assert domain.nodes[0] - f_min_new < domain.delta_f[0]
    assert domain.nodes[-1] <= f_max_new
    assert f_max_new - domain.nodes[-1] < domain.delta_f[-1]
    # Test that the length is correct.
    len_new = 0
    nodes_new = domain.nodes
    delta_f_initial_new = domain.delta_f[0]
    for n in range(1, len(nodes_new)):
        len_new += int(
            (nodes_new[n] - nodes_new[n - 1]) / delta_f_initial_new / 2 ** (n - 1)
        )
    assert len_new == len(domain())
    assert len_new == len(domain)
    # Test that the boundaries are correct
    mfd_original = MultibandedFrequencyDomain(**mfd_params)
    len_original = len(mfd_original)
    assert len_original == domain._range_update_initial_length
    assert (
        domain._range_update_idx_upper - domain._range_update_idx_lower + 1 == len_new
    )
    idx_lower = 0
    while mfd_original.sample_frequencies[idx_lower] < f_min_new:
        idx_lower += 1
    assert idx_lower == domain._range_update_idx_lower
    idx_upper = mfd_original.max_idx
    while f_max_new < mfd_original.sample_frequencies[idx_upper]:
        idx_upper -= 1
    assert idx_upper == domain._range_update_idx_upper
    # Test new mask
    assert len(domain.frequency_mask) == len_new
    assert np.all(domain.frequency_mask == 1.0)
    # Test that data range adjustment works
    a = np.random.random(len_original)
    assert np.all(
        a[domain._range_update_idx_lower : domain._range_update_idx_upper + 1]
        == domain.update_data(a)
    )
    # Test that data range adjustment works specifying different axis
    a = np.random.random((20, len_original, 30))
    assert np.all(
        a[:, domain._range_update_idx_lower : domain._range_update_idx_upper + 1, :]
        == domain.update_data(a, axis=1)
    )


def test_mfd_base_domain_consistency(mfd):
    assert np.all(mfd.sample_frequencies == mfd.decimate(mfd.base_domain()))
    assert mfd.f_min == mfd.base_domain.f_min
    assert mfd.f_max in mfd.base_domain.sample_frequencies


def test_mfd_time_translation(mfd):
    domain = mfd
    # test normal time translation
    dt = 1e-3
    a = np.sin(np.outer(np.arange(3) + 1, domain()) / 100) + 1j * np.cos(
        np.outer(np.arange(3) + 1, domain()) / 100
    )
    a_dt = a * np.exp(-2j * np.pi * domain() * dt)
    b_dt = domain.time_translate_data(a, dt)
    assert np.allclose(a_dt, b_dt)
    assert not np.allclose(a_dt, domain.time_translate_data(a, 1.01 * dt))
    # test that time translation can be undone
    assert np.allclose(a, domain.time_translate_data(a_dt, -dt))
    assert not np.allclose(a, domain.time_translate_data(a_dt, -1.01 * dt))


def test_mfd_time_translation_torch(mfd):
    domain = mfd
    batch_size = 5
    num_detectors = 2
    dt = torch.randn(batch_size, num_detectors, dtype=torch.float32)
    data = torch.empty(
        (batch_size, num_detectors, 3, len(domain) - domain.min_idx),
        dtype=torch.float32,
    )
    constant_value = 2.0
    f = domain.sample_frequencies_torch[domain.min_idx :]
    for i in range(batch_size):
        for j in range(num_detectors):
            signal = torch.exp(-2j * np.pi * dt[i, j] * f)
            data[i, j, 0, :] = signal.real
            data[i, j, 1, :] = signal.imag
            data[i, j, 2, :] = constant_value
    result = domain.time_translate_data(data, -dt)
    assert result.dtype == torch.float32
    assert result.shape == data.shape
    assert torch.allclose(result[..., 0, :], torch.tensor(1.0), atol=1e-4)
    # Tolerance of 1e-2 is required in the imaginary part, likely because we are
    # checking that it *vanishes*. This is a consequence of single-precision floats.
    # TODO: Is there a way to improve on this?
    assert torch.allclose(result[..., 1, :], torch.tensor(0.0), atol=1e-2)
    assert torch.allclose(result[..., 2, :], torch.tensor(constant_value))




import os 
from bilby_pipe.job_creation.nodes import PESummaryNode as BilbyPESummaryNode
from bilby_pipe.utils import BilbyPipeError, logger

logger.name = "dingo_pipe"


class PESummaryNode(BilbyPESummaryNode):

    run_node_on_osg = False
    
    def __init__(self, inputs, merged_node_list, generation_node_list, dag):
        super(BilbyPESummaryNode, self).__init__(inputs)
        self.dag = dag
        self.job_name = f"{inputs.label}_pesummary"
        self.request_cpus = 1

        n_results = len(merged_node_list)
        result_files = [merged_node.result_file for merged_node in merged_node_list]
        labels = [merged_node.label for merged_node in merged_node_list]

        if self.inputs.transfer_files or self.inputs.osg:
            input_files_to_transfer = [
                self._relative_topdir(fname, self.inputs.initialdir)
                for fname in result_files
            ] + inputs.additional_transfer_paths
            files = [self.inputs.complete_ini_file]
            if len(generation_node_list) == 1:
                files.append(generation_node_list[0].data_dump_file)
            input_files_to_transfer.extend(
                [
                    self._relative_topdir(fname, self.inputs.initialdir)
                    for fname in files
                ]
            )
            for value in [
                self.inputs.psd_dict,
                self.inputs.spline_calibration_envelope_dict,
            ]:
                input_files_to_transfer.extend(self.extract_paths_from_dict(value))
            if self.transfer_container:
                input_files_to_transfer.append(self.inputs.container)

            input_files_to_transfer, need_scitokens = self.job_needs_authentication(
                input_files_to_transfer
            )
            if need_scitokens:
                self.extra_lines.extend(self.scitoken_lines)

            self.extra_lines.extend(
                self._condor_file_transfer_lines(
                    input_files_to_transfer,
                    [self._relative_topdir(self.inputs.webdir, self.inputs.initialdir)],
                )
            )

        self.setup_arguments(
            add_ini=False, add_unknown_args=False, add_command_line_args=False
        )
        self.arguments.add("webdir", self.inputs.webdir)
        if self.inputs.email is not None:
            self.arguments.add("email", self.inputs.email)
        self.arguments.add(
            "config", " ".join([self.inputs.complete_ini_file] * n_results)
        )
        self.arguments.add("samples", f"{' '.join(result_files)}")

        # Using append here as summary pages doesn't take a full name for approximant
        self.arguments.append("-a")
        self.arguments.append(" ".join([self.inputs.waveform_approximant] * n_results))

        # adding the psds which were output as part of the data generation
        self.arguments.append("--psd")
        psd_str = " ".join([f"{det}:{os.path.join(self.inputs.outdir, 'data', f'{det}_psd.txt')}" for det in self.inputs.detectors])
        self.arguments.append(psd_str)

        # if len(generation_node_list) == 1:
        #     self.arguments.add("gwdata", generation_node_list[0].data_dump_file)
        # elif len(generation_node_list) > 1:
        #     logger.info(
        #         "Not adding --gwdata to PESummary job as there are multiple files"
        #     )
        existing_dir = self.inputs.existing_dir
        if existing_dir is not None:
            self.arguments.add("existing_webdir", existing_dir)

        if isinstance(self.inputs.summarypages_arguments, dict):
            if "labels" not in self.inputs.summarypages_arguments.keys():
                self.arguments.append(f"--labels {' '.join(labels)}")
            else:
                if len(labels) != len(result_files):
                    raise BilbyPipeError(
                        "Please provide the same number of labels for postprocessing "
                        "as result files"
                    )
            not_recognised_arguments = {}
            for key, val in self.inputs.summarypages_arguments.items():
                if key == "nsamples_for_skymap":
                    self.arguments.add("nsamples_for_skymap", val)
                elif key == "gw":
                    self.arguments.add_flag("gw")
                elif key == "no_ligo_skymap":
                    self.arguments.add_flag("no_ligo_skymap")
                elif key == "burnin":
                    self.arguments.add("burnin", val)
                elif key == "kde_plot":
                    self.arguments.add_flag("kde_plot")
                elif key == "gracedb":
                    self.arguments.add("gracedb", val)
                elif key == "palette":
                    self.arguments.add("palette", val)
                elif key == "include_prior":
                    self.arguments.add_flag("include_prior")
                elif key == "notes":
                    self.arguments.add("notes", val)
                elif key == "publication":
                    self.arguments.add_flag("publication")
                elif key == "labels":
                    self.arguments.add("labels", f"{' '.join(val)}")
                else:
                    not_recognised_arguments[key] = val
            if not_recognised_arguments != {}:
                logger.warn(
                    "Did not recognise the summarypages_arguments {}. To find "
                    "the full list of available arguments, please run "
                    "summarypages --help".format(not_recognised_arguments)
                )

        self.process_node()
        for merged_node in merged_node_list:
            self.job.add_parent(merged_node.job)



"""
This tests the method WaveformGenerator.generate_hplus_hcross_m, that returns the
polarzations disentangled into contributions m \in [-l_max, ...,0, ...,l_max],
that transform as exp(-1j * m * phase_shift) under phase shifts. This is important when
treating the phase parameter as an extrinsic parameter.

Note: this only accounts for the modified argument in the spherical harmonics, not for
the rotation of phase_shift of the cartesian spins in xy plane. Our workaround is to
set wfg.spin_conversion_phase = 0.0, which sets a constant phase 0 when converting PE
spins to cartesian spins. This means that phi_12 and phi_jl have different definitions,
which needs to be accounted for in postprocessing. The tests below all use
wfg.spin_conversion_phase = 0.0.
"""
import pytest
import numpy as np
from matplotlib import pyplot as plt

from dingo.gw.waveform_generator import (
    WaveformGenerator,
    sum_contributions_m,
    NewInterfaceWaveformGenerator,
)
from dingo.gw.gwutils import get_mismatch
from dingo.gw.domains import build_domain
from dingo.gw.prior import build_prior_with_defaults


@pytest.fixture
def uniform_fd_domain():
    domain_settings = {
        "type": "UniformFrequencyDomain",
        "f_min": 10.0,
        "f_max": 2048.0,  # Note that if this isn't a power of 2, mismatches are worse.
        "delta_f": 0.125,
    }
    domain = build_domain(domain_settings)
    return domain


@pytest.fixture(params=["IMRPhenomXPHM", "SEOBNRv4PHM", "SEOBNRv5PHM", "SEOBNRv5HM"])
def approximant(request):
    return request.param


@pytest.fixture
def intrinsic_prior(approximant):
    if "PHM" in approximant:
        intrinsic_dict = {
            "mass_1": "bilby.core.prior.Constraint(minimum=10.0, maximum=80.0)",
            "mass_2": "bilby.core.prior.Constraint(minimum=10.0, maximum=80.0)",
            "mass_ratio": "bilby.gw.prior.UniformInComponentsMassRatio(minimum=0.125, maximum=1.0)",
            "chirp_mass": "bilby.gw.prior.UniformInComponentsChirpMass(minimum=25.0, maximum=100.0)",
            "luminosity_distance": 1000.0,
            "theta_jn": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
            "phase": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
            "a_1": "bilby.core.prior.Uniform(minimum=0.0, maximum=0.99)",
            "a_2": "bilby.core.prior.Uniform(minimum=0.0, maximum=0.99)",
            "tilt_1": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
            "tilt_2": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
            "phi_12": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
            "phi_jl": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
            "geocent_time": 0.0,
        }
    else:
        # Aligned spins
        intrinsic_dict = {
            "mass_1": "bilby.core.prior.Constraint(minimum=10.0, maximum=80.0)",
            "mass_2": "bilby.core.prior.Constraint(minimum=10.0, maximum=80.0)",
            "mass_ratio": "bilby.gw.prior.UniformInComponentsMassRatio(minimum=0.125, maximum=1.0)",
            "chirp_mass": "bilby.gw.prior.UniformInComponentsChirpMass(minimum=25.0, maximum=100.0)",
            "luminosity_distance": 1000.0,
            "theta_jn": "bilby.core.prior.Sine(minimum=0.0, maximum=np.pi)",
            "phase": 'bilby.core.prior.Uniform(minimum=0.0, maximum=2*np.pi, boundary="periodic")',
            "chi_1": 'bilby.gw.prior.AlignedSpin(name="chi_1", a_prior=Uniform(minimum=0, maximum=0.99))',
            "chi_2": 'bilby.gw.prior.AlignedSpin(name="chi_2", a_prior=Uniform(minimum=0, maximum=0.99))',
            "geocent_time": 0.0,
        }
    prior = build_prior_with_defaults(intrinsic_dict)
    return prior


@pytest.fixture
def wfg(uniform_fd_domain, approximant):
    if approximant in ["SEOBNRv5PHM", "SEOBNRv5HM"]:
        wfg_class = NewInterfaceWaveformGenerator
    else:
        wfg_class = WaveformGenerator
    return wfg_class(
        approximant=approximant,
        domain=uniform_fd_domain,
        f_ref=10.0,
        f_start=10.0,
        spin_conversion_phase=0.0,
    )


@pytest.fixture
def num_evaluations(approximant):
    if "Phenom" in approximant:
        return 10
    elif approximant == "SEOBNRv4PHM":
        return 1
    else:
        return 10


@pytest.fixture
def tolerances(approximant):
    # Return (max, median) mismatches expected.
    if approximant == "IMRPhenomXPHM":
        # The mismatches are typically be of order 1e-5 to 1e-9. This comes from the
        # calculation of the magnitude of the orbital angular momentum, which we calculate
        # to a different order the IMRPhenomXPHM. It's tricky to get this exactly right,
        # since there are many different methods for this. But the small mismatches we do
        # get should not have a big effect in practice.
        return 2e-2, 1e-5

    elif approximant == "SEOBNRv4PHM":
        # The mismatches are typically be of order 1e-5. This is exclusively due to
        # different tapering. The reference polarizations are tapered and FFTed on the
        # level of polarizations, while for generate_hplus_hcross_m, the tapering and FFT
        # happens on the level of complex modes.
        # We tested the mismatches for 20k waveforms, and the largest mismatch encountered
        # was 7e-4, while almost all mismatches were of order 1e-5.
        return 5e-4, 5e-4

    elif approximant in ["SEOBNRv5PHM", "SEOBNRv5HM"]:
        # Tested on 1000 mismatches.
        return 1e-9, 1e-12

    else:
        return 1e-5, 1e-5


# Uncomment to test only one approximant.
try:
    import pyseobnr

    approximant_list = ["IMRPhenomXPHM", "SEOBNRv4PHM", "SEOBNRv5PHM", "SEOBNRv5HM"]
except ImportError:
    approximant_list = ["IMRPhenomXPHM", "SEOBNRv4PHM"]


@pytest.mark.parametrize("approximant", approximant_list)
def test_generate_hplus_hcross_m(intrinsic_prior, wfg, num_evaluations, tolerances):
    mismatches = []
    for idx in range(num_evaluations):
        p = intrinsic_prior.sample()
        phase_shift = np.random.uniform(high=2 * np.pi)

        pol_m = wfg.generate_hplus_hcross_m(p)
        pol = sum_contributions_m(pol_m, phase_shift=phase_shift)
        pol_ref = wfg.generate_hplus_hcross({**p, "phase": p["phase"] + phase_shift})

        mismatches.append(
            [
                get_mismatch(
                    pol[pol_name],
                    pol_ref[pol_name],
                    wfg.domain,
                    asd_file="aLIGO_ZERO_DET_high_P_asd.txt",
                )
                for pol_name in pol
            ]
        )

        debug = False
        if debug:
            maxval = max(mismatches[-1])
            idx = mismatches[-1].index(maxval)
            p = list(pol.keys())[idx]
            plt.figure(figsize=(10, 7))
            plt.plot(wfg.domain.sample_frequencies, pol[p], label="reconstructed")
            plt.plot(
                wfg.domain.sample_frequencies, pol_ref[p], label="ref", linestyle="--"
            )
            plt.plot(wfg.domain.sample_frequencies, pol_ref[p] - pol[p], label="diff")
            plt.legend()
            plt.xscale("log")
            plt.xlim((5, 128))
            plt.title(f"{p}, mismatch={maxval}")
            plt.show()

    mismatches = np.array(mismatches)

    assert np.max(mismatches) < tolerances[0]
    assert np.median(mismatches) < tolerances[1]



import inspect
import textwrap

import torch
from torch import nn

from .cflow_base import ContinuousFlowPosteriorModel


class FlowMatchingPosteriorModel(ContinuousFlowPosteriorModel):
    __doc__ = (
        inspect.getdoc(ContinuousFlowPosteriorModel)
        + "\n\n"
        + textwrap.dedent(
            """\
        For flow matching, the vector field represents the velocity vector field 
        for a particle trajectory. Training proceeds as follows:

            t               ~ U[0, 1-eps)                               noise level
            theta_0         ~ N(0, 1)                                   sampled noise
            theta_1         = theta                                     pure sample
            theta_t         = c1(t) * theta_1 + c0(t) * theta_0         noisy sample
        
            eps             = 0
            c0              = (1 - (1 - sigma_min) * t)
            c1              = t
        
            v_target        = theta_1 - (1 - sigma_min) * theta_0
            loss            = || v_target - network(theta_t, t) ||
            """
        )
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.eps = 0
        self.sigma_min = self.model_kwargs["posterior_kwargs"]["sigma_min"]

    def evaluate_vector_field(self, t, theta_t, *context_data):
        """
        Evaluate the vector field v(t, theta_t, context_data) that generates the flow
        via the ODE

            d/dt f(theta_t, t, context) = v(f(theta_t, t, context), t, context).

        For flow matching, the vector field is regressed directly during training.

        Parameters
        ----------
        t: float
            time (noise level)
        theta_t: torch.Tensor
            noisy parameters, perturbed with noise level t
        *context_data: list[torch.tensor]
            list with context data (GW data)
        """
        # If t is a number (and thus the same for each element in this batch),
        # expand it as a tensor. This is required for the odeint solver.
        t = t * torch.ones(len(theta_t), device=theta_t.device)
        return self.network(t, theta_t, *context_data)

    def loss(self, theta, *context):
        """
        Calculates loss as the mean squared error between the predicted vector field and
        the vector field for transporting the parameter data to samples from the prior.

        Parameters
        ----------
        theta: torch.Tensor
            Parameter values at which to evaluate the density. Should have a batch
            dimension (even if size B = 1).
        context: torch.Tensor
            Context information (typically observed data). Must have the same leading
            (batch) dimension as theta.

        Returns
        -------
        loss: torch.Tensor
            Mean loss across the batch (a scalar).
        """
        # Shall we allow for multiple time evaluations for every data, context pair (to improve efficiency)?
        mse = nn.MSELoss()

        t = self.sample_t(len(theta))
        theta_0 = self.sample_theta_0(len(theta))
        theta_1 = theta
        theta_t = ot_conditional_flow(theta_0, theta_1, t, self.sigma_min)
        true_vf = theta - (1 - self.sigma_min) * theta_0

        predicted_vf = self.network(t, theta_t, *context)
        loss = mse(predicted_vf, true_vf)
        return loss


def ot_conditional_flow(x_0, x_1, t, sigma_min):
    return (1 - (1 - sigma_min) * t)[:, None] * x_0 + t[:, None] * x_1



# adapted from the Asimov Bilby Pipeline interface
import configparser
import glob
import importlib
import os
import re
import subprocess
import time


from asimov import config, logger

from asimov.pipeline import (
    Pipeline,
    PipelineException,
    PipelineLogger,
    PESummaryPipeline,
)


class Dingo(Pipeline):
    """
    The Dingo Pipeline.

    Parameters
    ----------
    production : :class:`asimov.Production`
       The production object.
    category : str, optional
        The category of the job.
        Defaults to "C01_offline".
    """

    with importlib.resources.path("dingo.asimov", "dingo.ini") as template_file:
        config_template = template_file

    name = "dingo"
    STATUS = {"wait", "stuck", "stopped", "running", "finished"}

    def __init__(self, production, category=None):
        super(Dingo, self).__init__(production, category)
        self.logger = logger
        if not production.pipeline.lower() == self.name:
            raise PipelineException

    def detect_completion(self):
        """
        Check for the production of the posterior file to signal that the job has completed.
        """
        self.logger.info("Checking if the dingo job has completed")

        results_dir = glob.glob(f"{self.production.rundir}/result")
        if len(results_dir) > 0:
            if (
                len(
                    glob.glob(
                        os.path.join(results_dir[0], f"*importance_sampling.hdf5")
                    )
                )
                > 0
            ):
                self.logger.info("Results files found, the job is finished.")
                return True
            else:
                self.logger.info("No results files found.")
                return False
        else:
            self.logger.info("No results directory found")
            return False

    def before_submit(self):
        """Pre-submit hook."""
        self.logger.info("Running the before_submit hook")
        pass

    def build_dag(self, psds=None, user=None, clobber_psd=False, dryrun=False):
        """
        Construct a DAG file in order to submit a production to the
        condor scheduler using DingoPipe.

        Parameters
        ----------
        production : str
           The production name.
        psds : dict, optional
           The PSDs which should be used for this DAG. If no PSDs are
           provided the PSD files specified in the ini file will be used
           instead.
        user : str
           The user accounting tag which should be used to run the job.
        dryrun: bool
           If set to true the commands will not be run, but will be printed to standard output. Defaults to False.



        Raises
        ------
        PipelineException
           Raised if the construction of the DAG fails.
        """

        cwd = os.getcwd()
        self.logger.info(f"Working in {cwd}")

        if self.production.event.repository:
            ini = self.production.event.repository.find_prods(
                self.production.name, self.category
            )[0]
            ini = os.path.join(cwd, ini)
        else:
            ini = f"{self.production.name}.ini"

        rundir = self.production.rundir

        if "job label" in self.production.meta:
            job_label = self.production.meta["job label"]
        else:
            job_label = self.production.name

        command = [
            os.path.join(config.get("pipelines", "environment"), "bin", "dingo_pipe"),
            ini,
        ]

        if dryrun:
            print(" ".join(command))
        else:
            self.logger.info(" ".join(command))
            pipe = subprocess.Popen(
                command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )
            out, err = pipe.communicate()
            self.logger.info(out)

            if err or "DAG generation complete, to submit jobs" not in str(out):
                self.logger.error(err)
                raise PipelineException(
                    f"DAG file could not be created.\n{command}\n{out}\n\n{err}",
                    production=self.production.name,
                )

            else:
                time.sleep(10)
                return PipelineLogger(message=out, production=self.production.name)

    def samples(self):
        """
        Collect the combined samples files for PESummary.
        """

        results_dir = os.path.join(self.production.rundir, "result")
        results_filenames = glob.glob(
            os.path.join(results_dir, f"*importance_sampling.hdf5")
        )
        return os.path.join(results_dir, results_filenames[0])

    def upload_assets(self):
        """
        Upload the samples from this job.
        """

        asset = self.collect_assets()["samples"]
        self.production.event.repository.add_file(
            asset,
            "posterior_samples.hdf5",
            commit_message=f"Added posterior_samples.",
        )

    def collect_assets(self):
        """
        Gather all of the results assets for this job.
        """
        return {"samples": self.samples()}

    def collect_logs(self):
        """
        Collect all of the log files which have been produced by this production and
        return their contents as a dictionary.
        """
        logs = glob.glob(f"{self.production.rundir}/submit/*.err") + glob.glob(
            f"{self.production.rundir}/log*/*.err"
        )
        logs += glob.glob(f"{self.production.rundir}/*/*.out")
        messages = {}
        for log in logs:
            try:
                with open(log, "r") as log_f:
                    message = log_f.read()
                    message = message.split("\n")
                    messages[log.split("/")[-1]] = "\n".join(message[-100:])
            except FileNotFoundError:
                messages[log.split("/")[-1]] = (
                    "There was a problem opening this log file."
                )
        return messages

    def submit_dag(self, dryrun=False):
        """
        Submit a DAG file to the condor cluster.

        Parameters
        ----------
        dryrun : bool
           If set to true the DAG will not be submitted,
           but all commands will be printed to standard
           output instead. Defaults to False.

        Returns
        -------
        int
           The cluster ID assigned to the running DAG file.
        PipelineLogger
           The pipeline logger message.

        Raises
        ------
        PipelineException
           This will be raised if the pipeline fails to submit the job.
        """

        cwd = os.getcwd()
        self.logger.info(f"Working in {cwd}")

        self.before_submit()

        try:
            if "job label" in self.production.meta:
                job_label = self.production.meta["job label"]
            else:
                job_label = self.production.name
            dag_filename = f"dag_{job_label}.submit"
            command = [
                "condor_submit_dag",
                "-batch-name",
                f"dingo/{self.production.event.name}/{self.production.name}",
                os.path.join(self.production.rundir, "submit", dag_filename),
            ]

            if dryrun:
                print(" ".join(command))
            else:
                dagman = subprocess.Popen(
                    command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                )

                self.logger.info(" ".join(command))

                stdout, stderr = dagman.communicate()

                if "submitted to cluster" in str(stdout):
                    cluster = re.search(
                        r"submitted to cluster ([\d]+)", str(stdout)
                    ).groups()[0]
                    self.logger.info(
                        f"Submitted successfully. Running with job id {int(cluster)}"
                    )
                    self.production.status = "running"
                    self.production.job_id = int(cluster)
                    return cluster, PipelineLogger(stdout)
                else:
                    self.logger.error("Could not submit the job to the cluster")
                    self.logger.info(stdout)
                    self.logger.error(stderr)

                    raise PipelineException(
                        "The DAG file could not be submitted.",
                    )

        except FileNotFoundError as error:
            self.logger.exception(error)
            raise PipelineException(
                "It looks like condor isn't installed on this system.\n"
                f"""I wanted to run {" ".join(command)}."""
            ) from error

    def after_completion(self):
        post_pipeline = PESummaryPipeline(production=self.production)
        self.logger.info("Job has completed. Running PE Summary.")
        cluster = post_pipeline.submit_dag()
        self.production.meta["job id"] = int(cluster)
        self.production.status = "processing"
        self.production.event.update_data()

    def detect_completion_processing(self):
        # no post processing currently performed
        return True

    def resurrect(self):
        """
        Attempt to ressurrect a failed job.
        """
        try:
            count = self.production.meta["resurrections"]
        except:
            count = 0
        if (count < 5) and (
            len(glob.glob(os.path.join(self.production.rundir, "submit", "*.rescue*")))
            > 0
        ):
            count += 1
            self.submit_dag()

    @classmethod
    def read_ini(cls, filepath):
        """
        Read and parse a dingo configuration file.

        Parameters
        ----------
        filepath: str
           The path to the ini file.
        """

        with open(filepath, "r") as f:
            file_content = f.read()

        config_parser = configparser.RawConfigParser()
        config_parser.read_string(file_content)

        return config_parser



import argparse
import os.path
import pickle
from io import StringIO
from os.path import join
from os.path import isfile
import glob
import copy

import numpy as np
import requests
import yaml

from gwpy.table import EventTable
from dingo.gw.noise.asd_dataset import ASDDataset

"""
Catalogue against which to check that no event is present in the estimated PSDs
"""
CATALOGS = ["GWTC-1-confident", "GWTC-2.1-confident", "GWTC-3-confident"]

"""
Contains links for PSD segment lists with quality label BURST_CAT2 from the Gravitational Wave Open Science Center.
Some events are split up into multiple chunks such that there are multiple URLs for one observing run
"""
URL_DIRECTORY = {
    "O1_L1": [
        "https://www.gw-openscience.org/timeline/segments/O1/L1_BURST_CAT2/1126051217/1137254417/"
    ],
    "O1_H1": [
        "https://www.gw-openscience.org/timeline/segments/O1/H1_BURST_CAT2/1126051217/1137254417/"
    ],
    "O2_L1": [
        "https://www.gw-openscience.org/timeline/segments/O2_16KHZ_R1/L1_BURST_CAT2/1164556817/1187733618/"
    ],
    "O2_H1": [
        "https://www.gw-openscience.org/timeline/segments/O2_16KHZ_R1/H1_BURST_CAT2/1164556817/1187733618/"
    ],
    "O2_V1": [
        "https://www.gw-openscience.org/timeline/segments/O2_16KHZ_R1/V1_BURST_CAT2/1164556817/1187733618/"
    ],
    "O3_L1": [
        "https://www.gw-openscience.org/timeline/segments/O3a_16KHZ_R1/L1_BURST_CAT2/1238166018/1253977218/",
        "https://www.gw-openscience.org/timeline/segments/O3b_16KHZ_R1/L1_BURST_CAT2/1256655618/1269363618/",
    ],
    "O3_H1": [
        "https://www.gw-openscience.org/timeline/segments/O3a_16KHZ_R1/H1_BURST_CAT2/1238166018/1253977218/",
        "https://www.gw-openscience.org/timeline/segments/O3b_16KHZ_R1/H1_BURST_CAT2/1256655618/1269363618/",
    ],
    "O3_V1": [
        "https://www.gw-openscience.org/timeline/segments/O3a_16KHZ_R1/V1_BURST_CAT2/1238166018/1253977218/",
        "https://www.gw-openscience.org/timeline/segments/O3b_16KHZ_R1/V1_BURST_CAT2/1256655618/1269363618/",
    ],
}


def psd_data_path(data_dir, run, detector):
    """
    Return the directory where the PSD data is to be stored
    Parameters
    ----------
    data_dir : str
        Path to the directory where the PSD dataset will be stored
    run : str
        Observing run that is used for the PSD dataset generation
    detector : str
        Detector that is used for the PSD dataset generation
    Returns
    -------
    the path where the data is stored
    """
    return join(data_dir, "tmp", run, detector)


def get_event_gps_times():
    event_list = []
    for catalogue in CATALOGS:
        events = EventTable.fetch_open_data(catalogue)
        event_list += list(events["GPS"])
    return event_list


def get_time_segments(settings):
    """
    Creates a dictionary storing time segments used for estimating PSDs
    Parameters
    ----------
    settings : dict
        Settings that determine the segments
    Returns
    -------
    Dictionary containing the time segments for each detector
    """

    time_segments = {}

    run = settings["observing_run"]
    time_psd = settings["time_psd"]
    time_gap = settings["time_gap"]
    num_psds_max = settings.get("num_psds_max")

    event_list = get_event_gps_times()

    for detector in settings["detectors"]:

        key = run + "_" + detector
        urls = URL_DIRECTORY[key]

        starts, stops, durations = [], [], []
        for url in urls:
            r = requests.get(url, allow_redirects=True)
            c = StringIO(r.content.decode("utf-8"))
            starts_seg, stops_seg, durations_seg = np.loadtxt(
                c, dtype="int", unpack=True
            )
            starts = np.hstack([starts, starts_seg])
            stops = np.hstack([stops, stops_seg])
            durations = np.hstack([durations, durations_seg])

        segs = np.array(list(zip(starts, stops, durations)), dtype=int)
        segs = segs[segs[:, 2] >= time_psd]

        valid_segments = []
        for idx in range(segs.shape[0]):
            seg = segs[idx, :]
            start_time = seg[0]
            end_time = start_time + time_psd

            while end_time <= seg[1]:
                # check that no event is present in the segments
                if not any(start_time <= event <= end_time for event in event_list):
                    valid_segments.append((start_time, end_time))
                start_time = end_time + time_gap
                end_time = start_time + time_psd

        if num_psds_max is not None and num_psds_max > 0:
            valid_segments = valid_segments[:num_psds_max]

        time_segments[detector] = valid_segments

    return time_segments


def merge_datasets(asd_dataset_list):
    """
    Merges a list of asd datasets into ont
    Parameters
    ----------
    asd_dataset_list: List of ASDDatasets to be merged

    Returns
    -------
    A single combined ASDDataset object
    """

    merged_dict = {"asds": {}, "gps_times": {}}

    for det, asd_list in asd_dataset_list.items():
        print(f"Merging {len(asd_list)} datasets into one for detector {det}.")
        merged_dict["asds"][det] = np.vstack(
            [asd_dataset.asds[det] for asd_dataset in asd_list]
        )
        merged_dict["gps_times"][det] = np.hstack(
            [asd_dataset.gps_times[det] for asd_dataset in asd_list]
        )

    # copy settings from last dataset
    merged_dict["settings"] = copy.deepcopy(asd_list[-1].settings)

    merged = ASDDataset(dictionary=merged_dict)
    return merged


def merge_datasets_cli():
    """
    Command-line function to combine a collection of datasets into one. Used for
    parallelized ASD dataset generation.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data_dir",
        type=str,
        required=True,
        help="Path where the PSD data is to be stored.",
    )
    parser.add_argument(
        "--settings_file",
        type=str,
        required=True,
        help="Path to a settings file that contains the settings for the ASD dataset generation",
    )
    parser.add_argument(
        "--time_segments_file",
        type=str,
        default=None,
        help="Path to a file containing the time segments for which PSDs should be estimated",
    )
    parser.add_argument(
        "--num_parts",
        type=int,
        default=-1,
        help="Number of ASD datasets that should be merged",
    )
    parser.add_argument(
        "--out_name",
        type=str,
        default=None,
        help="File name of merged dataset",
    )
    args = parser.parse_args()

    with open(args.settings_file, "r") as f:
        settings = yaml.safe_load(f)

    time_segments = None
    if args.time_segments_file:
        with open(args.time_segments_file, "rb") as f:
            time_segments = pickle.load(f)

    detectors = settings["dataset_settings"]["detectors"]
    observing_run = settings["dataset_settings"]["observing_run"]
    asd_dataset_list = {det: [] for det in detectors}

    for det in detectors:
        file_dir = psd_data_path(args.data_dir, observing_run, det)
        if time_segments:
            filenames = [
                join(file_dir, f"asd_{seg[0]}.hdf5")
                for seg in time_segments[det]
                if isfile(join(file_dir, f"asd_{seg[0]}.hdf5"))
            ]
        else:  # if no time_segments are specified, use the first 'num_parts' ASD datasets
            filenames = sorted(glob.glob(join(file_dir, f"asd_*.hdf5")))
            num_parts = min(args.num_parts, len(filenames)) if args.num_parts > 0 else len(filenames)
            filenames = filenames[:num_parts]

        asd_dataset_list[det] = [ASDDataset(filename) for filename in filenames]

    merged_dataset = merge_datasets(asd_dataset_list)
    filename = args.out_name
    if filename is None:
        run = settings["dataset_settings"]["observing_run"]
        filename = join(args.data_dir, f"asds_{run}.hdf5")
    merged_dataset.to_file(filename)



import pytest
import random
import numpy as np
import torch
from bilby.gw.detector import InterferometerList

from dingo.gw.transforms import time_delay_from_geocenter
from dingo.gw.prior import BBHExtrinsicPriorDict, default_extrinsic_dict


@pytest.fixture
def ifo_list_and_prior_and_tref():
    ifo_list = InterferometerList(["H1", "L1", "V1"])
    prior = BBHExtrinsicPriorDict({k: default_extrinsic_dict[k] for k in ["ra", "dec"]})
    tref = random.randint(
        946339215, 1577491218
    )  # random gps time between 2010 and 2030
    return ifo_list, prior, tref


def test_time_delay_from_geocenter_errors(ifo_list_and_prior_and_tref):
    ifo_list, prior, tref = ifo_list_and_prior_and_tref
    params = prior.sample(1000)
    with pytest.raises(ValueError):
        ra = params["ra"]
        dec = torch.Tensor(params["dec"].item())
        time_delay_from_geocenter(ifo_list[0], ra, dec, tref)
    with pytest.raises(ValueError):
        ra = params["ra"]
        dec = params["dec"][:100]
        time_delay_from_geocenter(ifo_list[0], ra, dec, tref)
    with pytest.raises(ValueError):
        ra = params["ra"][None,:]
        dec = params["dec"][None,:]
        time_delay_from_geocenter(ifo_list[0], ra, dec, tref)


def test_time_delay_from_geocenter_float(ifo_list_and_prior_and_tref):
    ifo_list, prior, tref = ifo_list_and_prior_and_tref
    params = prior.sample(1)
    ra = params["ra"].item()
    dec = params["dec"].item()
    for ifo in ifo_list:
        dt0 = ifo.time_delay_from_geocenter(ra, dec, tref)
        dt1 = time_delay_from_geocenter(ifo, ra, dec, tref)
        assert np.allclose(dt0, dt1, atol=1e-10)


def test_time_delay_from_geocenter_array(ifo_list_and_prior_and_tref):
    ifo_list, prior, tref = ifo_list_and_prior_and_tref
    params = prior.sample(1000)
    # check that our vectorized function returns the same result as the bilby method
    ra = params["ra"]
    dec = params["dec"]
    for ifo in ifo_list:
        dt0 = [
            ifo.time_delay_from_geocenter(ra_idx, dec_idx, tref)
            for ra_idx, dec_idx in zip(ra, dec)
        ]
        dt0 = np.array(dt0)
        dt1 = time_delay_from_geocenter(ifo, ra, dec, tref)
        assert np.allclose(dt0, dt1, atol=1e-10)
    # check that this also works for arrays of length 1, for which the bilby method is
    # called since it is faster
    ra = params["ra"][:1]
    dec = params["dec"][:1]
    for ifo in ifo_list:
        dt0 = [
            ifo.time_delay_from_geocenter(ra_idx, dec_idx, tref)
            for ra_idx, dec_idx in zip(ra, dec)
        ]
        dt0 = np.array(dt0)
        dt1 = time_delay_from_geocenter(ifo, ra, dec, tref)
        assert dt0 == dt1


def test_time_delay_from_geocenter_tensor(ifo_list_and_prior_and_tref):
    ifo_list, prior, tref = ifo_list_and_prior_and_tref
    params = prior.sample(1000)
    # check that our vectorized function returns the same result as the bilby method
    ra = torch.Tensor(params["ra"])
    dec = torch.Tensor(params["dec"])
    for ifo in ifo_list:
        dt0 = [
            ifo.time_delay_from_geocenter(ra_idx, dec_idx, tref)
            for ra_idx, dec_idx in zip(ra, dec)
        ]
        dt0 = np.array(dt0)
        dt1 = time_delay_from_geocenter(ifo, ra, dec, tref)
        assert np.allclose(dt0, dt1, atol=1e-6) # larger atol due to single precision
    # check that this also works for arrays of length 1, for which the bilby method is
    # called since it is faster
    ra = torch.Tensor(params["ra"])[:1]
    dec = torch.Tensor(params["dec"])[:1]
    for ifo in ifo_list:
        dt0 = [
            ifo.time_delay_from_geocenter(ra_idx, dec_idx, tref)
            for ra_idx, dec_idx in zip(ra, dec)
        ]
        dt0 = np.array(dt0)
        dt1 = time_delay_from_geocenter(ifo, ra, dec, tref)
        assert dt0 == dt1



from dingo.core.posterior_models.base_model import BasePosteriorModel
from dingo.core.posterior_models.normalizing_flow import NormalizingFlowPosteriorModel
from dingo.core.posterior_models.cflow_base import ContinuousFlowPosteriorModel
from dingo.core.posterior_models.flow_matching import FlowMatchingPosteriorModel
from dingo.core.posterior_models.score_matching import ScoreDiffusionPosteriorModel



import numpy as np
from functools import lru_cache

from .base import Domain


class TimeDomain(Domain):
    """Defines the physical time domain on which the data of interest live.

    The time bins are assumed to be uniform between [0, duration]
    with spacing 1 / sampling_rate.
    """

    def __init__(self, time_duration: float, sampling_rate: float):
        self._time_duration = time_duration
        self._sampling_rate = sampling_rate

    @lru_cache()
    def __len__(self):
        """Number of time bins given duration and sampling rate"""
        return int(self._time_duration * self._sampling_rate)

    @lru_cache()
    def __call__(self) -> np.ndarray:
        """Array of uniform times at which data is sampled"""
        num_bins = self.__len__()
        return np.linspace(
            0.0, self._time_duration, num=num_bins, endpoint=False, dtype=np.float32
        )

    @property
    def delta_t(self) -> float:
        """The size of the time bins"""
        return 1.0 / self._sampling_rate

    @delta_t.setter
    def delta_t(self, delta_t: float):
        self._sampling_rate = 1.0 / delta_t

    @property
    def noise_std(self) -> float:
        """Standard deviation of the whitened noise distribution.

        To have noise that comes from a multivariate *unit* normal
        distribution, you must divide by this factor. In practice, this means
        dividing the whitened waveforms by this.

        In the continuum limit in time domain, the standard deviation of white
        noise would at each point go to infinity, hence the delta_t factor.
        """
        return 1.0 / np.sqrt(2.0 * self.delta_t)

    def time_translate_data(self, data, dt) -> np.ndarray:
        raise NotImplementedError

    @property
    def f_max(self) -> float:
        """The maximum frequency [Hz] is typically set to half the sampling
        rate."""
        return self._sampling_rate / 2.0

    @property
    def duration(self) -> float:
        """Waveform duration in seconds."""
        return self._time_duration

    @property
    def sampling_rate(self) -> float:
        return self._sampling_rate

    @property
    def min_idx(self) -> int:
        return 0

    @property
    def max_idx(self) -> int:
        return round(self._time_duration * self._sampling_rate)

    @property
    def domain_dict(self):
        """Enables to rebuild the domain via calling build_domain(domain_dict)."""
        return {
            "type": "TimeDomain",
            "time_duration": self._time_duration,
            "sampling_rate": self._sampling_rate,
        }


import numpy as np
import torch
import torch.nn as nn
from torch.nn import functional as F
from torch.utils.data import DataLoader
from typing import Union, Tuple, Iterable
import bilby


def fix_random_seeds(_):
    """Utility function to set random seeds when using multiple workers for DataLoader."""
    np.random.seed(int(torch.initial_seed()) % (2 ** 32 - 1))
    try:
        bilby.core.utils.random.seed(int(torch.initial_seed()) % (2 ** 32 - 1))
    except AttributeError:  # In case using an old version of Bilby.
        pass


def get_activation_function_from_string(activation_name: str):
    """
    Returns an activation function, based on the name provided.

    :param activation_name: str
        name of the activation function, one of {'elu', 'relu', 'leaky_rely'}
    :return: function
        corresponding activation function
    """
    if activation_name.lower() == "elu":
        return F.elu
    elif activation_name.lower() == "relu":
        return F.relu
    elif activation_name.lower() == "leaky_relu":
        return F.leaky_relu
    elif activation_name.lower() == "gelu":
        return F.gelu
    else:
        raise ValueError("Invalid activation function.")


def get_number_of_model_parameters(
    model: nn.Module,
    requires_grad_flags: tuple = (True, False),
):
    """
    Counts parameters of the module. The list requires_grad_flag can be used
    to specify whether all parameters should be counted, or only those with
    requires_grad = True or False.
    :param model: nn.Module
        model
    :param requires_grad_flags: tuple
        tuple of bools, for requested requires_grad flags
    :return:
        number of parameters of the model with requested required_grad flags
    """
    num_params = 0
    for p in list(model.parameters()):
        if p.requires_grad in requires_grad_flags:
            n = 1
            for s in list(p.size()):
                n = n * s
            num_params += n
    return num_params


def get_optimizer_from_kwargs(
    model_parameters: Iterable,
    **optimizer_kwargs,
):
    """
    Builds and returns an optimizer for model_parameters. The type of the
    optimizer is determined by kwarg type, the remaining kwargs are passed to
    the optimizer.

    Parameters
    ----------
    model_parameters: Iterable
        iterable of parameters to optimize or dicts defining parameter groups
    optimizer_kwargs:
        kwargs for optimizer; type needs to be one of [adagrad, adam, adamw,
        lbfgs, RMSprop, sgd], the remaining kwargs are used for specific
        optimizer kwargs, such as learning rate and momentum

    Returns
    -------
    optimizer
    """
    optimizers_dict = {
        "adagrad": torch.optim.Adagrad,
        "adam": torch.optim.Adam,
        "adamw": torch.optim.AdamW,
        "lbfgs": torch.optim.LBFGS,
        "RMSprop": torch.optim.RMSprop,
        "sgd": torch.optim.SGD,
    }
    if not "type" in optimizer_kwargs:
        raise KeyError("Optimizer type needs to be specified.")
    if not optimizer_kwargs["type"].lower() in optimizers_dict:
        raise ValueError("No valid optimizer specified.")
    optimizer = optimizers_dict[optimizer_kwargs.pop("type")]
    return optimizer(model_parameters, **optimizer_kwargs)


def get_scheduler_from_kwargs(
    optimizer: torch.optim.Optimizer,
    **scheduler_kwargs,
):
    """
    Builds and returns an scheduler for optimizer. The type of the
    scheduler is determined by kwarg type, the remaining kwargs are passed to
    the scheduler.

    Parameters
    ----------
    optimizer: torch.optim.optimizer.Optimizer
        optimizer for which the scheduler is used
    scheduler_kwargs:
        kwargs for scheduler; type needs to be one of [step, cosine,
        reduce_on_plateau], the remaining kwargs are used for
        specific scheduler kwargs, such as learning rate and momentum

    Returns
    -------
    scheduler
    """
    schedulers_dict = {
        "step": torch.optim.lr_scheduler.StepLR,
        "cosine": torch.optim.lr_scheduler.CosineAnnealingLR,
        "reduce_on_plateau": torch.optim.lr_scheduler.ReduceLROnPlateau,
    }
    if not "type" in scheduler_kwargs:
        raise KeyError("Scheduler type needs to be specified.")
    if not scheduler_kwargs["type"].lower() in schedulers_dict:
        raise ValueError("No valid scheduler specified.")
    scheduler = schedulers_dict[scheduler_kwargs.pop("type")]
    return scheduler(optimizer, **scheduler_kwargs)


def perform_scheduler_step(
    scheduler,
    loss=None,
):
    """
    Wrapper for scheduler.step(). If scheduler is ReduceLROnPlateau,
    then scheduler.step(loss) is called, if not, scheduler.step().

    Parameters
    ----------

    scheduler:
        scheduler for learning rate
    loss:
        validation loss
    """
    if type(scheduler) == torch.optim.lr_scheduler.ReduceLROnPlateau:
        scheduler.step(loss)
    else:
        scheduler.step()


def get_lr(optimizer):
    """Returns a list with the learning rates of the optimizer."""
    return [param_group["lr"] for param_group in optimizer.state_dict()["param_groups"]]


def split_dataset_into_train_and_test(dataset, train_fraction):
    """
    Splits dataset into a trainset of size int(train_fraction * len(dataset)),
    and a testset with the remainder. Uses fixed random seed for
    reproducibility.

    Parameters
    ----------
    dataset: torch.utils.data.Datset
        dataset to be split
    train_fraction: float
        fraction of the dataset to be used for trainset

    Returns
    -------
    trainset, testset
    """
    train_size = int(train_fraction * len(dataset))
    test_size = len(dataset) - train_size
    return torch.utils.data.random_split(
        dataset, [train_size, test_size], generator=torch.Generator().manual_seed(42)
    )


def build_train_and_test_loaders(
    dataset: torch.utils.data.Dataset,
    train_fraction: float,
    batch_size: int,
    num_workers: int,
):
    """
    Split the dataset into train and test sets, and build corresponding DataLoaders.
    The random split uses a fixed seed for reproducibility.

    Parameters
    ----------
    dataset : torch.utils.data.Dataset
    train_fraction : float
        Fraction of dataset to use for training. The remainder is used for testing.
        Should lie between 0 and 1.
    batch_size : int
    num_workers : int

    Returns
    -------
    (train_loader, test_loader)
    """

    # Split the dataset. This function uses a fixed seed for reproducibility.
    train_dataset, test_dataset = split_dataset_into_train_and_test(
        dataset, train_fraction
    )

    # Build DataLoaders
    train_loader = DataLoader(
        train_dataset,
        batch_size=batch_size,
        shuffle=True,
        pin_memory=True,
        num_workers=num_workers,
        worker_init_fn=fix_random_seeds,
    )
    test_loader = DataLoader(
        test_dataset,
        batch_size=batch_size,
        shuffle=False,
        pin_memory=True,
        num_workers=num_workers,
        worker_init_fn=fix_random_seeds,
    )

    return train_loader, test_loader


def set_requires_grad_flag(
    model, name_startswith=None, name_contains=None, requires_grad=True
):
    """
    Set param.requires_grad of all model parameters with a name starting with
    name_startswith, or name containing name_contains, to requires_grad.
    """
    for name, param in model.named_parameters():
        if (
            name_startswith is not None
            and name.startswith(name_startswith)
            or name_contains is not None
            and name_contains in name
        ):
            param.requires_grad = requires_grad


def torch_detach_to_cpu(x):
    if isinstance(x, torch.Tensor):
        return x.detach().cpu()
    return x
